"""Utilities for working with json_schema."""
from typing import Any, Dict


def json_schema_type(in_type: str) -> str:
    if in_type == "long_string":
        return "string"
    return in_type


def create_entry_json_schema(fields: Dict[str, Any]):
    """Create json_schema from fields definition.

    Args:
        fields (Dict[str, Any]): the fields config to process

    Returns:
        Dict[str]: The json_schema to use.
    """
    json_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {},
    }

    def recursive_field(parent_schema, parent_field_name, parent_field_def):
        if parent_field_def.get("virtual", False):
            return

        if parent_field_def["type"] != "object":
            # TODO this will not work when we have user defined types, s.a. saldoid
            schema_type = json_schema_type(parent_field_def["type"])
            result = {"type": schema_type}
        else:
            result = {"type": "object", "properties": {}}

            for child_field_name, child_field_def in parent_field_def["fields"].items():
                recursive_field(result, child_field_name, child_field_def)

        if parent_field_def.get("required", False):
            if "required" not in parent_schema:
                parent_schema["required"] = []
            parent_schema["required"].append(parent_field_name)

        if parent_field_def.get("collection", False):
            result = {"type": "array", "items": result}

        parent_schema["properties"][parent_field_name] = result

    for field_name, field_def in fields.items():
        recursive_field(json_schema, field_name, field_def)

    return json_schema



from typing import Dict, List
import uuid

import pytest  # pyre-ignore
from fastapi import status

from karp import auth
from karp.errors import ClientErrorCodes
from karp.lex.application.queries.resources import GetEntryRepositoryId
from karp.lex.application.repositories.entry_repositories import (
    EntryUowRepositoryUnitOfWork,
)
from karp.foundation.time import utc_now
from karp.foundation.value_objects import make_unique_id, UniqueId, unique_id
from karp.lex.application.queries import EntryDto

# from karp.application import ctx, config
# from karp.infrastructure.unit_of_work import unit_of_work


# from tests.utils import get_json

# from tests.integration_tests.common_fixtures import (
#     fixture_fa_data_client,
#     fixture_places,
# )


def get_entry_uow(container, resource_id: str):
    get_entry_repository_id = container.get(GetEntryRepositoryId)  # type: ignore [misc]
    entry_repo_id = get_entry_repository_id.query("places")
    entry_uow_repo_uow = container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
    with entry_uow_repo_uow as uw:
        return uw.repo.get_by_id(entry_repo_id)


def init(
    client,
    entries: List[Dict],
    access_token: auth.AccessToken,
) -> list[str]:

    result = []
    for entry in entries:
        response = client.post(
            "/entries/places/add",
            json={"entry": entry},
            headers=access_token.as_header(),
        )
        assert response.status_code < 300, response.status_code
        result.append(response.json()["newID"])
    return result


@pytest.fixture(name="entry_places_214_id", scope="session")
def fixture_entry_places_214_id(
    fa_data_client,
    write_token: auth.AccessToken,
):

    ids = init(
        fa_data_client,
        [
            {"code": 214, "name": "last_modified1", "municipality": [1]},
        ],
        write_token,
    )
    return ids[0]


@pytest.fixture(name="entry_places_209_id", scope="session")
def fixture_entry_places_209_id(
    fa_data_client,
    write_token: auth.AccessToken,
):

    ids = init(
        fa_data_client,
        [
            {"code": 209, "name": "a", "municipality": [1]},
        ],
        write_token,
    )
    return ids[0]


class TestEntriesRoutes:
    def test_routes_exist(self, fa_data_client):
        response = fa_data_client.post("/entries/places/add")
        assert response.status_code != status.HTTP_404_NOT_FOUND

        response = fa_data_client.post("/entries/places/update")
        assert response.status_code != status.HTTP_404_NOT_FOUND

        response = fa_data_client.put("/entries/places")
        assert response.status_code != status.HTTP_404_NOT_FOUND

        response = fa_data_client.post("/entries/places/preview")
        assert response.status_code != status.HTTP_404_NOT_FOUND


class TestAddEntry:
    def test_put_route_exist(self, fa_data_client):
        response = fa_data_client.put("/entries/places")
        assert response.status_code != status.HTTP_404_NOT_FOUND

    @pytest.mark.parametrize(
        "invalid_data",
        [
            ({},),
            ({"user": "a@b.se"},),
        ],
    )
    def test_invalid_data_returns_422(
        self,
        fa_data_client,
        invalid_data: Dict,
        write_token: auth.AccessToken,
    ):
        response = fa_data_client.post(
            "/entries/places/add",
            json=invalid_data,
            headers=write_token.as_header(),
        )
        print(f"{response.json()=}")
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    def test_add_with_valid_data_returns_201(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):

        response = fa_data_client.put(
            "/entries/places",
            json={
                "entry": {
                    "code": 203,
                    "name": "add203",
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f"response. = {response.json()}")
        assert response.status_code == 201
        response_data = response.json()
        assert "newID" in response_data
        new_id = unique_id.parse(response_data["newID"])

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        with entry_uow as uw:
            assert new_id in uw.repo.entity_ids()

    def test_add_with_valid_data_and_entity_id_returns_201(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entity_id = make_unique_id()
        response = fa_data_client.put(
            "/entries/places",
            json={
                "entry": {
                    "code": 2078,
                    "name": "add203",
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "id": str(entity_id),
            },
            headers=write_token.as_header(),
        )
        print(f"response. = {response.json()}")
        assert response.status_code == 201
        response_data = response.json()
        assert "newID" in response_data
        new_id = response_data["newID"]

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        with entry_uow as uw:
            assert new_id in uw.repo.entity_ids()

    @pytest.mark.skip(reason="we don't use entry_id")
    def test_adding_existing_fails_with_400(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 204
        entry_name = f"add{entry_id}"
        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f"response = {response.json()}")

        assert response.status_code == 201

        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == 400
        response_data = response.json()

        assert "error" in response_data
        assert "errorCode" in response_data
        assert ClientErrorCodes.DB_INTEGRITY_ERROR == response_data["errorCode"]
        assert (
            response_data["error"]
            == f"An entry with entry_id '{entry_id}' already exists."
        )

    def test_add_fails_with_invalid_entry(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        response = fa_data_client.post(
            "/entries/places/add", json={"entry": {}}, headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_400_BAD_REQUEST
        response_data = response.json()
        assert response_data["errorCode"] == ClientErrorCodes.ENTRY_NOT_VALID

        # assert (
        #     response_data["error"] == "Missing ID field for resource 'places' in '{}'"
        # )


class TestDeleteEntry:
    def test_delete(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 205
        entry_name = f"delete{entry_id}"
        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )

        entity_id = response.json()["newID"]
        response = fa_data_client.delete(
            f"/entries/places/{entity_id}/delete", headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_204_NO_CONTENT

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        with entry_uow as uw:
            assert uw.repo.by_id(entity_id).discarded
            assert entity_id not in uw.repo.entity_ids()

    def test_delete_rest(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 205
        entry_name = f"delete{entry_id}"
        response = fa_data_client.put(
            "/entries/places",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )

        entity_id = response.json()["newID"]
        response = fa_data_client.delete(
            f"/entries/places/{entity_id}", headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_204_NO_CONTENT

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        with entry_uow as uw:
            assert uw.repo.by_id(entity_id).discarded
            assert entity_id not in uw.repo.entity_ids()

    def test_delete_non_existing_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):

        entry_id = make_unique_id()

        response = fa_data_client.delete(
            f"/entries/places/{entry_id}/delete", headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND

        response_data = response.json()

        assert response_data["errorCode"] == ClientErrorCodes.ENTRY_NOT_FOUND

        assert (
            response_data["error"]
            == f"Entry '{entry_id}' not found in resource 'places' (version=latest)"
        )

    def test_delete_rest_non_existing_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):

        entry_id = "00000000000000000000000000"

        response = fa_data_client.delete(
            f"/entries/places/{entry_id}", headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND

        response_data = response.json()

        assert response_data["errorCode"] == ClientErrorCodes.ENTRY_NOT_FOUND

        assert (
            response_data["error"]
            == f"Entry '{entry_id}' not found in resource 'places' (version=latest)"
        )


class TestUpdateEntry:
    def test_update_non_existing_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = make_unique_id()
        response = fa_data_client.post(
            f"/entries/places/{entry_id}/update",
            json={
                "entry": {
                    "code": 3,
                    "name": "test3",
                    "population": 5,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "message": "changes",
                "version": 1,
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND
        response_data = response.json()
        assert response_data["errorCode"] == ClientErrorCodes.ENTRY_NOT_FOUND

        assert (
            response_data["error"]
            == f"Entry '{entry_id}' not found in resource 'places' (version=latest)"
        )

    def test_update_wo_changes_succeeds(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 206
        entry_name = f"update{entry_id}"
        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f"response = {response.json()}")
        entity_id = response.json()["newID"]
        assert response.status_code == status.HTTP_201_CREATED

        response = fa_data_client.post(
            f"/entries/places/{entity_id}/update",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "message": "changes",
                "version": 1,
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK

    def test_update_wrong_version_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 207
        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": "update3",
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f"response = {response.json()}")
        entity_id = response.json()["newID"]

        assert response.status_code == 201

        response = fa_data_client.post(
            f"/entries/places/{entity_id}/update",
            json={
                "entry": {
                    "code": entry_id,
                    "name": "update3",
                    "population": 5,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "message": "changes",
                "version": 2,
            },
            headers=write_token.as_header(),
            #         ),
            #         content_type='application/json',
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        response_data = response.json()
        assert response_data["errorCode"] == ClientErrorCodes.VERSION_CONFLICT

        assert response_data["error"] == "Version conflict. Please update entry."
        assert response_data["diff"] == [
            {"type": "CHANGE", "field": "population", "before": 4, "after": 5}
        ]

    def test_update_returns_200_on_valid_data(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 208
        entry_name = f"update{entry_id}"
        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f"response = {response.json()}")
        entity_id = response.json()["newID"]

        assert response.status_code == status.HTTP_201_CREATED

        response = fa_data_client.post(
            f"/entries/places/{entity_id}/update",
            json={
                "entry": {
                    "code": entry_id,
                    "name": entry_name,
                    "population": 5,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "message": "changes",
                "version": 1,
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        assert response_data["newID"] == entity_id

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        with entry_uow as uw:
            assert uw.repo.by_id(entity_id).body["population"] == 5
            assert entity_id in uw.repo.entity_ids()

    def test_update_several_times(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
        entry_places_209_id: str,
    ):
        entry_id = 209

        entity_id = entry_places_209_id

        for i in range(2, 10):
            response = fa_data_client.post(
                f"/entries/places/{entity_id}/update",
                json={
                    "entry": {"code": entry_id, "name": "a" * i, "municipality": [1]},
                    "message": "changes",
                    "version": i - 1,
                },
                headers=write_token.as_header(),
            )
            print(f"i = {i}: response = {response.json()}")
            assert response.status_code == status.HTTP_200_OK

    @pytest.mark.xfail()
    def test_update_entry_id(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 210
        response = fa_data_client.post(
            "/entries/places/add",
            json={
                "entry": {
                    "code": entry_id,
                    "name": "update3",
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == 201

        response = fa_data_client.post(
            f"places/{entry_id}/update",
            json={
                "entry": {
                    "code": entry_id + 1,
                    "name": "update3",
                    "population": 5,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "message": "changes",
                "version": 1,
            },
            headers=write_token.as_header(),
        )
        response_data = response.json()
        print(f"{response.json()=}")
        assert str(entry_id + 1) == response_data["newID"]

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        with entry_uow as uw:
            entry_ids = uw.repo.entry_ids()
            assert str(entry_id) not in entry_ids
            assert str(entry_id + 1) in entry_ids

    def test_update_changes_last_modified(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 212
        before_add = utc_now()

        ids = init(
            fa_data_client,
            [
                {"code": entry_id, "name": "last_modified1", "municipality": [1]},
            ],
            write_token,
        )

        after_add = utc_now()

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container, resource_id="places"
        )
        entity_id = ids[0]
        with entry_uow as uw:
            entry = uw.repo.by_id(entity_id)
            assert entry.last_modified > before_add
            assert entry.last_modified < after_add

        fa_data_client.post(
            f"/entries/places/{entity_id}/update",
            json={
                "entry": {
                    "code": entry_id,
                    "name": "last_modified2",
                    "municipality": [1],
                },
                "message": "changes",
                "version": 1,
            },
            headers=write_token.as_header(),
        )

        after_update = utc_now()

        with entry_uow as uw:
            entry = uw.repo.by_id(entity_id)
            assert entry.last_modified > after_add
            assert entry.last_modified < after_update


class TestGetEntry:
    def test_get_entry_wo_auth_returns_403(
        self, fa_data_client, entry_places_214_id: str
    ):
        response = fa_data_client.get(f"/entries/places/{entry_places_214_id}")
        assert response.status_code == status.HTTP_403_FORBIDDEN

    def test_get_entry_w_lower_auth_returns_401(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
        entry_places_214_id: str,
    ):
        response = fa_data_client.get(
            f"/entries/places/{entry_places_214_id}",
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_401_UNAUTHORIZED

    def test_get_entry_by_entry_id(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        entry_places_214_id: str,
    ):
        response = fa_data_client.get(
            f"/entries/places/{entry_places_214_id}",
            headers=admin_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK

        entry = EntryDto(**response.json())
        assert entry.entity_id == entry_places_214_id
        assert entry.version == 1

    def test_route_w_version_exist(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        entry_places_209_id: str,
    ):
        response = fa_data_client.get(
            f"/entries/places/{entry_places_209_id}/5",
            headers=admin_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK

        entry = EntryDto(**response.json())
        assert entry.entity_id == entry_places_209_id
        assert entry.version == 5


class TestPreviewEntry:
    def test_preview_fails_with_422_on_invalid_data(
        self, fa_data_client, read_token: auth.AccessToken
    ):
        response = fa_data_client.post(
            "/entries/places/preview", json={}, headers=read_token.as_header()
        )
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    def test_preview_returns_200_on_valid_data(
        self, fa_data_client, read_token: auth.AccessToken
    ):
        response = fa_data_client.post(
            "/entries/places/preview",
            json={
                "entry": {
                    "code": 3,
                    "name": "update3",
                    "population": 4,
                    "area": 50000,
                    "density": 5,
                    "municipality": [2, 3],
                },
                "message": "test",
            },
            headers=read_token.as_header(),
        )
        print(f"{response.json()=}")
        assert response.status_code == status.HTTP_200_OK
        assert unique_id.parse(response.json()["entry"]["id"])


@pytest.mark.skip()
def test_update_wrong_id(fa_data_client, write_token: auth.AccessToken):
    response = fa_data_client.post(
        "/entries/places/add",
        json={
            "entry": {
                "code": 3,
                "name": "update3",
                "population": 4,
                "area": 50000,
                "density": 5,
                "municipality": [2, 3],
            }
        },
        headers=write_token.as_header(),
    )
    print(f"response = {response.json()}")

    assert response.status_code == 201
    entry_id = response.json()["newID"]

    with unit_of_work(using=ctx.resource_repo) as uw:
        resource = uw.get_active_resource("places")

    with unit_of_work(using=resource.entry_repository) as uw:
        entries = uw.entry_ids()
        assert len(entries) == 1
        assert entries[0] == "3"

    response = fa_data_client.post(
        f"places/{entry_id}/update",
        json={
            "entry": {
                "code": 4,
                "name": "update3",
                "population": 4,
                "area": 50000,
                "density": 5,
                "municipality": [2, 3],
            },
            "message": "changes",
            "version": 1,
        },
        headers=write_token.as_header(),
    )
    assert response.status_code == 400
    response_data = response.json()
    assert response_data["errorCode"] == ClientErrorCodes.ENTRY_ID_MISMATCH

    assert response_data["error"] == "entry_id '4' does not equal '3'"


@pytest.mark.skip()
def test_refs(fa_data_client):
    client = init(
        fa_data_client,
        [
            {
                "code": 1,
                "name": "refs1refs",
                "population": 10,
                "area": 50000,
                "density": 5,
                "municipality": [2, 3],
            },
            {
                "code": 2,
                "name": "refs2refs",
                "population": 5,
                "larger_place": 1,
                "area": 50000,
                "density": 5,
                "municipality": [2, 3],
            },
        ],
    )

    entries = get_json(client, "places/query")
    assert len(entries["hits"]) == 2
    for val in entries["hits"]:
        assert "entry" in val
        entry = val["entry"]
        print("entry = {}".format(entry))
        if entry["code"] == 1:
            assert "v_larger_place" not in entry
            assert "larger_place" not in entry
            assert "v_smaller_places" in entry
            assert entry["v_smaller_places"][0]["code"] == 2
        else:
            assert entry["v_larger_place"]["code"] == 1
            assert entry["v_larger_place"]["name"] == "refs1refs"
            assert "v_smaller_places" not in entry


@pytest.mark.skip()
def test_external_refs(fa_data_client):
    client = init(
        fa_data_client,
        [
            {
                "code": 1,
                "name": "test1",
                "population": 10,
                "area": 50000,
                "density": 5,
                "municipality": [1],
            },
            {
                "code": 2,
                "name": "test2",
                "population": 5,
                "larger_place": 1,
                "area": 50000,
                "density": 5,
                "municipality": [1, 2],
            },
            {
                "code": 3,
                "name": "test2",
                "population": 5,
                "larger_place": 1,
                "area": 50000,
                "density": 5,
                "municipality": [2],
            },
        ],
    )

    client.post(
        "municipalities/add",
        data=json.dumps(
            {
                "entry": {
                    "code": 1,
                    "name": "municipality1",
                    "state": "state1",
                    "region": "region1",
                }
            }
        ),
        content_type="application/json",
    )

    client.post(
        "municipalities/add",
        data=json.dumps(
            {
                "entry": {
                    "code": 2,
                    "name": "municipality2",
                    "state": "state2",
                    "region": "region2",
                }
            }
        ),
        content_type="application/json",
    )

    entries = get_json(client, "municipalities/query")
    for val in entries["hits"]:
        assert "entry" in val
        entry = val["entry"]

        assert "v_places" in entry
        place_codes = [place["code"] for place in entry["v_places"]]
        assert len(place_codes) == 2
        if entry["code"] == 1:
            assert 1 in place_codes
            assert 2 in place_codes
        else:
            assert 2 in place_codes
            assert 3 in place_codes

    places_entries = get_json(client, "places/query")
    for val in places_entries["hits"]:
        assert "entry" in val
        entry = val["entry"]
        assert "municipality" in entry
        assert isinstance(entry["v_municipality"], list)
        if entry["code"] == 2:
            assert {"code": 1, "name": "municipality1", "state": "state1"} in entry[
                "v_municipality"
            ]
            assert {"code": 2, "name": "municipality2", "state": "state2"} in entry[
                "v_municipality"
            ]


@pytest.mark.skip()
def test_update_refs(fa_data_client):
    client = init(
        fa_data_client,
        es,
        [
            {
                "code": 5,
                "name": "test1",
                "population": 10,
                "area": 50000,
                "density": 5,
                "municipality": [2, 3],
            },
            {
                "code": 6,
                "name": "test2",
                "population": 5,
                "larger_place": 5,
                "area": 50000,
                "density": 5,
                "municipality": [2, 3],
            },
        ],
    )

    entries = get_json(client, "places/query")
    assert len(entries["hits"]) == 2
    for val in entries["hits"]:
        assert "entry" in val
        entry = val["entry"]
        print("entry = {}".format(entry))
        if entry["code"] == 5:
            assert "v_smaller_places" in entry
            assert entry["v_smaller_places"][0]["code"] == 6

    client.delete("/places/6/delete")

    entries = get_json(client, "places/query")
    assert len(entries["hits"]) == 1
    entry = entries["hits"][0]
    assert "v_smaller_places" not in entry


@pytest.mark.skip()
def test_update_refs2(fa_data_client):
    client = init(
        fa_data_client, es, [{"code": 3, "name": "test3", "municipality": [2, 3]}]
    )

    client.post(
        "places/3/update",
        data=json.dumps(
            {
                "entry": {"code": 3, "name": "test3", "municipality": [2]},
                "message": "changes",
                "version": 1,
            }
        ),
        content_type="application/json",
    )

    entries = get_json(client, "places/query")
    assert len(entries["hits"]) == 1
    assert entries["hits"][0]["id"] == "3"
    assert entries["hits"][0]["entry"]["municipality"] == [2]
    assert (
        "v_municipality" not in entries["hits"][0]
        or len(entries["hits"][0]["municipality"]) == 0
    )
    with client.application.app_context():
        db_entry = entryread.get_entry("places", "3")
        assert len(db_entry.municipality) == 1
        assert db_entry.municipality[0].municipality == 2



import typing

from karp.auth.domain import errors
from karp.auth.application.queries import (
    GetResourcePermissions,
    ResourcePermissionDto,
    IsResourceProtected,
)
from karp.foundation.value_objects.permission_level import PermissionLevel
from karp.lex.application.queries import (
    GetPublishedResources,
    ReadOnlyResourceRepository,
)


class LexGetResourcePermissions(GetResourcePermissions):
    def __init__(
        self,
        get_published_resources: GetPublishedResources,
    ):
        self.get_published_resources = get_published_resources

    def query(self) -> typing.List[ResourcePermissionDto]:
        resource_permissions = []
        for resource in self.get_published_resources.query():
            resource_obj = {"resource_id": resource.resource_id}

            protected_conf = resource.config.get("protected")
            if not protected_conf:
                protected = None
            elif protected_conf.get("admin"):
                protected = "ADMIN"
            elif protected_conf.get("write"):
                protected = "WRITE"
            else:
                protected = "READ"

            if protected:
                resource_obj["protected"] = protected
            resource_permissions.append(resource_obj)

        return resource_permissions


class LexIsResourceProtected(IsResourceProtected):
    def __init__(self, resource_repo: ReadOnlyResourceRepository) -> None:
        super().__init__()
        self.resource_repo = resource_repo

    def query(self, resource_id: str, level: PermissionLevel) -> bool:
        if level in [PermissionLevel.write, PermissionLevel.admin]:
            return True
        resource = self.resource_repo.get_by_resource_id(resource_id=resource_id)
        if not resource:
            raise errors.ResourceNotFound(f"Can't find resource '{resource_id}'")
        return resource.config.get("protected", {}).get("read", False)



import unicodedata


def shave_marks(txt: str) -> str:
    """Remove all diacritic marks."""
    norm_txt = unicodedata.normalize("NFD", txt)
    shaved = "".join(c for c in norm_txt if not unicodedata.combining(c))
    return unicodedata.normalize("NFC", shaved)



"""Entity"""
import datetime
from typing import Optional, Union

from deprecated import deprecated

from karp.foundation.errors import ConsistencyError
from karp.foundation import errors
from karp.foundation import events
from karp.utility.time import monotonic_utc_now


class Entity(events.EventMixin):
    DiscardedEntityError = errors.DiscardedEntityError

    def __init__(self, entity_id, discarded: bool = False, aggregate_root=None):
        super().__init__()
        self._id = entity_id
        self._discarded = discarded
        self._root = aggregate_root

    def queue_event(self, event):
        self._record_event(event)

    @property
    def id(self):
        """A unique identifier for the entity."""
        return self._id

    @property
    def entity_id(self):
        """A unique identifier for the entity."""
        return self._id

    @property
    def discarded(self) -> bool:
        """True if this entity is marked as deleted, otherwise False."""
        return self._discarded

    @property
    def root(self):
        """The aggregate root or self."""
        return self if self._root is None else self._root

    def _check_not_discarded(self):
        if self._discarded:
            raise self.DiscardedEntityError(
                f"Attempt to use {self!r}, entity_id = {self.entity_id}"
            )

    def _validate_event_applicability(self, event):
        if event.entity_id != self.id:
            raise ConsistencyError(
                f"Event entity id mismatch: {event.entity_id} != {self.id}"
            )


class VersionedEntity(Entity):
    def __init__(self, entity_id, version: int, discarded: bool = False):
        super().__init__(entity_id, discarded=discarded)
        self._version = version

    @property
    def version(self) -> int:
        """An integer version for the entity."""
        return self._version

    def _increment_version(self):
        self._version += 1

    def _validate_version(self, version: int) -> None:
        if version != self.version:
            raise ConsistencyError(
                f"Entity version mismatch: {version} != {self.version}"
            )


class TimestampedEntity(Entity):
    def __init__(
        self,
        entity_id,
        last_modified: Optional[float] = None,
        last_modified_by: Optional[str] = None,
        discarded: bool = False,
    ) -> None:
        super().__init__(entity_id, discarded=discarded)
        self._last_modified = self._ensure_timestamp(last_modified)
        self._last_modified_by = (
            "Unknown user" if last_modified_by is None else last_modified_by
        )

    @property
    def last_modified(self):
        """The time this entity was last modified."""
        return self._last_modified

    @property
    def last_modified_by(self):
        """The time this entity was last modified."""
        return self._last_modified_by

    def _ensure_timestamp(
        self, timestamp: Optional[Union[float, datetime.datetime, str]]
    ) -> float:
        if isinstance(timestamp, datetime.datetime):
            return timestamp.timestamp()
        elif isinstance(timestamp, str):
            return datetime.datetime.fromisoformat(timestamp).timestamp()
        return monotonic_utc_now() if timestamp is None else timestamp

    def _validate_last_modified(self, last_modified: float):
        if int(last_modified) != int(self.last_modified):
            raise ConsistencyError(
                f"Event entity last_modified mismatch: {last_modified} != {self.last_modified}"
            )


class TimestampedVersionedEntity(VersionedEntity, TimestampedEntity):
    def __init__(
        self,
        entity_id,
        last_modified: Optional[float] = None,
        last_modified_by: Optional[str] = None,
        discarded: bool = False,
        *,
        version: int,
    ) -> None:
        super().__init__(entity_id, version=version, discarded=discarded)
        TimestampedEntity.__init__(
            self,
            entity_id=entity_id,
            discarded=discarded,
            last_modified=last_modified,
            last_modified_by=last_modified_by,
        )



import logging
import typing

from karp.foundation.value_objects.permission_level import PermissionLevel

logger = logging.getLogger(__name__)


class User:
    def __init__(
        self,
        identifier: str,
        permissions: typing.Dict[str, int],
        levels: typing.Dict[str, int],
    ):
        self._identifier = identifier
        self._permissions = permissions
        self._levels = levels

    @property
    def identifier(self) -> str:
        return self._identifier

    def has_enough_permissions(self, resource_id: str, level: PermissionLevel) -> bool:
        try:
            return (self._permissions.get(resource_id) is not None) and (
                self._permissions[resource_id] >= self._levels[level]
            )
        except KeyError:
            logger.exception(
                "error checking permissions",
                extra={
                    "resource_id": resource_id,
                    "level": level,
                    "identifier": self._identifier,
                    "levels": self._levels,
                    "permissions": self._permissions,
                },
            )
            raise



from typing import List

import pytest

from karp.search.application.queries import QueryRequest
from karp.search.domain.query import Query
from karp.search_infrastructure.queries import EsQuery


@pytest.fixture()
def resource_ids() -> List[str]:
    return ["resource_a"]


@pytest.fixture()
def request_from() -> int:
    return 4


@pytest.fixture()
def request_size() -> int:
    return 5


@pytest.fixture()
def query_request(
    resource_ids: List[str], request_from: int, request_size: int
) -> QueryRequest:
    return QueryRequest(
        resource_ids=resource_ids, from_=request_from, size=request_size
    )


def test_create_EsQuery_from_QueryRequest(query_request: QueryRequest) -> None:
    query = EsQuery.from_query_request(query_request)

    assert query.from_ == query_request.from_
    assert query.size == query_request.size



import logging
import os
from distutils.util import strtobool
from pathlib import Path

from sqlalchemy.engine.url import URL, make_url
from starlette.config import Config as StarletteConfig
from starlette.datastructures import CommaSeparatedStrings, Secret

config = StarletteConfig(".env")

DEBUG = config("DEBUG", cast=bool, default=False)
TESTING = config("TESTING", cast=bool, default=False)

DB_DRIVER = config("DB_DRIVER", default="mysql+pymysql")
DB_HOST = config("DB_HOST", default=None)
DB_PORT = config("DB_PORT", cast=int, default=None)
DB_USER = config("DB_USER", default=None)
DB_PASSWORD = config("DB_PASSWORD", cast=str, default=None)
DB_DATABASE = config("DB_DATABASE", default=None)
DB_TEST_DATABASE = config("DB_TEST_DATABASE", default=None)
if TESTING is True:
    DB_DATABASE = DB_TEST_DATABASE or f"{DB_DATABASE}_test"
if DB_DRIVER != "sqlite" and not DB_HOST:
    DB_HOST = "localhost"
DB_URL = config(
    "DB_URL",
    cast=make_url,
    default=URL.create(
        drivername=DB_DRIVER,
        username=DB_USER,
        password=DB_PASSWORD,
        host=DB_HOST,
        port=DB_PORT,
        database=DB_DATABASE,
        query={"charset": "utf8mb4"},
    ),
)

ELASTICSEARCH_HOST = config(
    "ELASTICSEARCH_HOST", cast=CommaSeparatedStrings, default=None
)
# MYSQL_FORMAT = "mysql://{user}:{passwd}@{dbhost}/{dbname}"


# class Config:
#     SQLALCHEMY_TRACK_MODIFICATIONS = False
#     DEBUG = False
#     TESTING = False
#     SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
#     ELASTICSEARCH_HOST = (
#         os.environ["ELASTICSEARCH_HOST"].split(",")
#         if "ELASTICSEARCH_HOST" in os.environ
#         else None
#     )
TEST_ELASTICSEARCH_ENABLED = config(
    "TEST_ELASTICSEARCH_ENABLED", cast=bool, default=False
)

SEARCH_CONTEXT = config("SEARCH_CONTEXT", default=None)
AUTH_CONTEXT = config("AUTH_CONTEXT", default=None)

TEST_ES_HOME = config("TEST_ES_HOME", cast=Path, default=None)
#     # CONSOLE_LOG_LEVEL = logging.getLevelName(
#     #     os.environ.get("CONSOLE_LOG_LEVEL", "INFO")
#     # )


def get_loglevel(loglevel: str) -> int:
    return getattr(logging, loglevel.upper(), logging.WARNING)


CONSOLE_LOG_LEVEL = config("CONSOLE_LOG_LEVEL", cast=get_loglevel, default="INFO")
#     LOG_TO_SLACK = strtobool(os.environ.get("LOG_TO_SLACK", "n"))
#     SLACK_SECRET = os.environ.get("SLACK_SECRET")
#     JWT_AUTH = strtobool(os.environ.get("JWT_AUTH", "n"))

if TESTING:
    print("Using testing jwt key")
    JWT_AUTH_PUBKEY_PATH = Path(__file__).parent / "../tests/data/pubkey.pem"
else:
    JWT_AUTH_PUBKEY_PATH = config("JWT_AUTH_PUBKEY_PATH", cast=Path, default=None)

MYSQL_FORMAT = "mysql://{user}:{passwd}@{dbhost}/{dbname}"


class Config:
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    DEBUG = False
    TESTING = False
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    ELASTICSEARCH_HOST = (
        os.environ["ELASTICSEARCH_HOST"].split(",")
        if "ELASTICSEARCH_HOST" in os.environ
        else None
    )
    ELASTICSEARCH_ENABLED = strtobool(os.environ.get("ELASTICSEARCH_ENABLED", "n"))
    CONSOLE_LOG_LEVEL = getattr(
        logging, os.environ.get("CONSOLE_LOG_LEVEL", "INFO"), logging.INFO
    )
    LOG_TO_SLACK = strtobool(os.environ.get("LOG_TO_SLACK", "n"))
    SLACK_SECRET = os.environ.get("SLACK_SECRET")
    JWT_AUTH = strtobool(os.environ.get("JWT_AUTH", "n"))


class ProductionConfig(Config):
    def __init__(self):
        self.SQLALCHEMY_DATABASE_URI = MYSQL_FORMAT.format(
            user=os.environ["MARIADB_USER"],
            pwd=os.environ["MARIADB_PASSWORD"],
            dbhost=os.environ["MARIADB_HOST"],
            dbname=os.environ["MARIADB_DATABASE"],
        )


class DevelopmentConfig(Config):
    DEBUG = True


class TestingConfig(Config):
    TESTING = True


def get_config():
    return Config()


class MariaDBConfig(Config):
    def __init__(
        self, user=None, pwd=None, host=None, dbname=None, setup_database=False
    ):
        if not user:
            user = os.environ["MARIADB_USER"]
        if not pwd:
            pwd = os.environ["MARIADB_PASSWORD"]
        if not host:
            host = os.environ["MARIADB_HOST"]
        if not dbname:
            dbname = os.environ["MARIADB_DATABASE"]

        self.SETUP_DATABASE = setup_database
        self.SQLALCHEMY_DATABASE_URI = MYSQL_FORMAT.format(
            user=user, passwd=pwd, dbhost=host, dbname=dbname
        )



from karp import search
from .sql_search_service import SqlSearchService


class SqlSearchServiceUnitOfWork(search.IndexUnitOfWork):
    def __init__(self, session_factory):
        # session_factory=DEFAULT_SESSION_FACTORY):
        self.session_factory = session_factory
        self._index = SqlSearchService()

    def _commit(self):
        pass

    def rollback(self):
        pass

    @property
    def repo(self):
        if not self._index:
            raise RuntimeError()
        return self._index



import injector

from karp.foundation.commands import CommandHandler
from karp.lex.application.repositories import (
    EntryUowRepositoryUnitOfWork,
    EntryRepositoryUnitOfWorkFactory,
    InjectorEntryUnitOfWorkRepoFactory,
    ResourceUnitOfWork,
)
from karp.lex.domain.commands import (
    AddEntries,
    AddEntriesInChunks,
    AddEntry,
    CreateEntryRepository,
    CreateResource,
    DeleteEntryRepository,
    ImportEntries,
    ImportEntriesInChunks,
    SetEntryRepoId,
)
from karp.lex.domain import commands
from karp.lex.domain.commands.resource_commands import SetEntryRepoId
from karp.lex.domain.value_objects import EntrySchema
from karp.lex.application.use_cases import (
    AddingEntries,
    AddingEntry,
    CreatingEntryRepo,
    CreatingResource,
    DeletingEntry,
    DeletingEntryRepository,
    DeletingResource,
    ImportingEntries,
    PublishingResource,
    SettingEntryRepoId,
    UpdatingEntry,
    UpdatingResource,
)
from karp.lex.application.queries import (
    EntryDto,
    EntryViews,
    GetEntryDiff,
    GetEntryHistory,
    GetHistory,
    GetHistoryDto,
    GetPublishedResources,
    GetResources,
    HistoryDto,
    ResourceDto,
    ReadOnlyResourceRepository,
    EntryRepoDto,
    ReadOnlyEntryRepoRepositry,
    ListEntryRepos,
)


__all__ = [
    # commands
    "AddEntries",
    "AddEntriesInChunks",
    "ImportEntries",
    "ImportEntriesInChunks",
    # use cases
    "ImportingEntries",
]


class Lex(injector.Module):
    @injector.provider
    def entry_uow_factory(
        self, container: injector.Injector
    ) -> EntryRepositoryUnitOfWorkFactory:
        return InjectorEntryUnitOfWorkRepoFactory(container)

    @injector.provider
    def create_entry_repository(
        self,
        uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[CreateEntryRepository]:
        return CreatingEntryRepo(
            entry_repo_uow=uow,
        )

    @injector.provider
    def deleting_entry_repository(
        self,
        uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[DeleteEntryRepository]:
        return DeletingEntryRepository(
            entry_repo_uow=uow,
        )

    @injector.provider
    def create_resource(
        self,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[CreateResource]:
        return CreatingResource(
            entry_repo_uow=entry_repo_uow,
            resource_uow=resource_uow,
        )

    @injector.provider
    def deleting_resource(
        self,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[commands.DeleteResource]:
        return DeletingResource(resource_uow)

    @injector.provider
    def setting_entry_repo_id(
        self,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[SetEntryRepoId]:
        return SettingEntryRepoId(
            entry_repo_uow=entry_repo_uow,
            resource_uow=resource_uow,
        )

    @injector.provider
    def update_resource(
        self,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[commands.UpdateResource]:
        return UpdatingResource(resource_uow)

    @injector.provider
    def publish_resource(
        self,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[commands.PublishResource]:
        return PublishingResource(resource_uow)

    @injector.provider
    def add_entry(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[commands.AddEntry]:
        return AddingEntry(resource_uow=resource_uow, entry_repo_uow=entry_repo_uow)

    @injector.provider
    def add_entries(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[commands.AddEntries]:
        return AddingEntries(resource_uow=resource_uow, entry_repo_uow=entry_repo_uow)

    @injector.provider
    def adding_entries_in_chunks(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[AddEntriesInChunks]:
        return AddingEntries(resource_uow=resource_uow, entry_repo_uow=entry_repo_uow)

    @injector.provider
    def import_entries(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[commands.ImportEntries]:
        return ImportingEntries(
            resource_uow=resource_uow, entry_repo_uow=entry_repo_uow
        )

    @injector.provider
    def importing_entries_in_chunks(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[ImportEntriesInChunks]:
        return ImportingEntries(
            resource_uow=resource_uow, entry_repo_uow=entry_repo_uow
        )

    @injector.provider
    def update_entry(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[commands.UpdateEntry]:
        return UpdatingEntry(resource_uow=resource_uow, entry_repo_uow=entry_repo_uow)

    @injector.provider
    def delete_entry(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[commands.DeleteEntry]:
        return DeletingEntry(resource_uow=resource_uow, entry_repo_uow=entry_repo_uow)



# import json
import re
import time
from datetime import datetime, timezone

import pytest  # pyre-ignore
from starlette import status

from karp import auth
from karp.foundation.time import utc_now
from karp.lex.application.queries.entries import EntryDto, GetHistoryDto, HistoryDto

places = [
    {"code": 103, "name": "a", "municipality": [1]},
    {"code": 104, "name": "b", "municipality": [2, 3]},
    {"code": 105, "name": "c", "municipality": [2, 3], "larger_place": 4},
    {"code": 106, "name": "c", "municipality": [1], "larger_place": 5},
]


def get_helper(client, url: str, access_token: auth.AccessToken):
    response = client.get(
        url,
        headers=access_token.as_header(),
    )
    assert response.status_code == status.HTTP_200_OK
    return response.json()


def username(name: str) -> str:
    return f"{__name__}-{name}"


@pytest.fixture(name="history_entity_ids", scope="session")
def fixture_history_entity_ids(
    fa_data_client,
    user1_token: auth.AccessToken,
    user2_token: auth.AccessToken,
    user4_token: auth.AccessToken,
) -> list[str]:
    entity_ids = []
    for entry in places[:2]:
        response = fa_data_client.post(
            "/entries/places/add",
            json={"entry": entry, "message": "Add it"},
            headers=user1_token.as_header(),
        )
        print(f"{response.json()=}")
        assert response.status_code == status.HTTP_201_CREATED
        entity_ids.append(response.json()["newID"])
        time.sleep(1)

    for entry in places[2:]:
        response = fa_data_client.post(
            "/entries/places/add",
            json={"entry": entry, "message": "Add it"},
            headers=user2_token.as_header(),
        )
        assert response.status_code == status.HTTP_201_CREATED
        entity_ids.append(response.json()["newID"])
        time.sleep(1)

    for i, entry in enumerate(places[:2]):
        changed_entry = entry.copy()
        changed_entry["name"] = entry["name"] + entry["name"]
        fa_data_client.post(
            f"/entries/places/{entity_ids[i]}/update",
            json={"entry": changed_entry, "message": "change it", "version": 1},
            headers=user2_token.as_header(),
        )
        time.sleep(1)

    for i, entry in enumerate(places[2:], start=2):
        changed_entry = entry.copy()
        changed_entry["name"] = entry["name"] + entry["name"]
        fa_data_client.post(
            f"/entries/places/{entity_ids[i]}/update",
            json={"entry": changed_entry, "message": "change it", "version": 1},
            headers=user1_token.as_header(),
        )
        time.sleep(1)

    diff_entry_id = entity_ids[0]
    for i in range(3, 10):
        changed_entry = places[0].copy()
        changed_entry["name"] = places[0]["name"] * i
        response = fa_data_client.post(
            f"/entries/places/{diff_entry_id}/update",
            json={"entry": changed_entry, "message": "changes", "version": i - 1},
            headers=user4_token.as_header(),
        )
        print(f"response = {response.json()}")
        assert response.status_code == status.HTTP_200_OK
        time.sleep(1)
    return entity_ids


class TestGetHistory:
    def test_route_exist(self, fa_data_client):
        response = fa_data_client.get("/history/places")
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_empty_user_history(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client, "/history/places?user_id=user3", admin_token
        )
        assert len(response_data["history"]) == 0
        assert response_data["total"] == 0

    def test_user1_history(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            "/history/places?user_id=user1",
            admin_token,
        )
        assert len(response_data["history"]) == 4
        assert response_data["total"] == 4
        for history_entry in response_data["history"]:
            assert "user1" == history_entry["user_id"]

    def test_user2_history(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            "/history/places?user_id=user2",
            admin_token,
        )
        assert 4 == len(response_data["history"])
        assert response_data["history"][0]["op"] == "ADDED"
        assert response_data["history"][1]["op"] == "ADDED"
        assert "UPDATED" == response_data["history"][2]["op"]
        assert "UPDATED" == response_data["history"][3]["op"]
        assert re.match(
            r"^\d{10}\.\d{3,6}$", str(response_data["history"][3]["timestamp"])
        )
        for history_entry in response_data["history"]:
            assert "user2" == history_entry["user_id"]

    def test_user_history_from_date(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            f"/history/places?user_id=user1&from_date={utc_now() - 5}",
            admin_token,
        )
        assert 4 > len(response_data["history"]), len(response_data["history"])

    def test_user_history_to_date(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            f"/history/places?user_id=user2&to_date={utc_now() - 5}",
            admin_token,
        )
        assert len(response_data["history"]) == 4

    def test_entry_id(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            f"/history/places?entry_id={history_entity_ids[1]}",
            admin_token,
        )
        assert 2 == len(response_data["history"])
        for history_entry in response_data["history"]:
            assert history_entry["entity_id"] == history_entity_ids[1]

    def test_entry_id_and_user_id(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            f"/history/places?entry_id={history_entity_ids[0]}&user_id=user2",
            admin_token,
        )
        assert 1 == len(response_data["history"])
        history_entry = response_data["history"][0]
        assert history_entry["entity_id"] == history_entity_ids[0]
        assert "user2" == history_entry["user_id"]
        assert "UPDATED" == history_entry["op"]

    def test_diff_against_nothing(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response_data = get_helper(
            fa_data_client,
            f"/history/places?entry_id={history_entity_ids[0]}&to_version=2",
            admin_token,
        )
        assert 1 == len(response_data["history"])
        history_entry = response_data["history"][0]
        assert "ADDED" == history_entry["op"]
        for diff in history_entry["diff"]:
            print(f"diff = {diff}")
            assert "ADDED" == diff["type"]


def test_historical_entry(
    fa_data_client,
    admin_token: auth.AccessToken,
    history_entity_ids: list[str],
):
    response_data = get_helper(
        fa_data_client,
        f"/entries/places/{history_entity_ids[1]}/2",
        admin_token,
    )
    assert utc_now() > response_data["last_modified"]
    assert "user2" == response_data["last_modified_by"]
    assert "resource" in response_data
    assert "bb" == response_data["entry"]["name"]
    assert "version" in response_data
    history_entry = EntryDto(**response_data)
    assert history_entry.version == 2


class TestGetEntryDiff:
    def test_diff1(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_version=1&to_version=2",
            headers=user4_token.as_header(),
        )
        response_data = response.json()
        print(f"response_data = {response_data}")
        assert response.status_code == status.HTTP_200_OK

        diff = response_data["diff"]
        assert len(diff) == 1
        assert "CHANGE" == diff[0]["type"]
        assert "a" == diff[0]["before"]
        assert "aa" == diff[0]["after"]
        assert "name" == diff[0]["field"]
        assert response_data["from_version"] == 1
        assert response_data["to_version"] == 2

    def test_diff2(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_date=0&to_date={utc_now()}",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        print(f"response = {response_data}")
        assert "a" == diff[0]["before"]
        assert "aaaaaaaaa" == diff[0]["after"]
        assert response_data["from_version"] == 1
        assert response_data["to_version"] == 9

    def test_diff_from_first_to_date(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        """
        this test is a bit shaky due to assuming that we will find the correct version by subtracting three seconds
        """
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?to_date={utc_now() - 3}",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "a" == diff[0]["before"]
        assert diff[0]["after"] > "aaaaaa"
        assert response_data["from_version"] == 1
        assert response_data["to_version"] > 6

    @pytest.mark.xfail()
    def test_diff_from_date_to_last(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        """
        this test is a bit shaky due to assuming that we will find the correct version by subtracting three seconds
        """
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_date={utc_now() - 3}",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "aaaaaaaa" == diff[0]["before"]
        assert "aaaaaaaaa" == diff[0]["after"]
        assert response_data["from_version"] == 8
        assert response_data["to_version"] == 9

    def test_diff_from_first_to_version(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?to_version=7",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "a" == diff[0]["before"]
        assert "aaaaaaa" == diff[0]["after"]
        assert response_data["from_version"] == 1
        assert response_data["to_version"] == 7

    def test_diff_from_version_to_last(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_version=7",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "aaaaaaa" == diff[0]["before"]
        assert "aaaaaaaaa" == diff[0]["after"]
        assert response_data["from_version"] == 7
        assert response_data["to_version"] == 9

    def test_diff_mix_version_date(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_version=2&to_date={utc_now() - 3}",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "aa" == diff[0]["before"]
        assert diff[0]["after"] > "aaaaaa"
        assert response_data["from_version"] == 2
        assert response_data["to_version"] > 6

    def test_diff_to_entry_data(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        edited_entry = places[0].copy()
        edited_entry["name"] = "testing"
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_version=1",
            json=edited_entry,
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "a" == diff[0]["before"]
        assert "testing" == diff[0]["after"]
        assert response_data["from_version"] == 1
        assert response_data["to_version"] is None

    def test_diff_no_flags(
        self,
        fa_data_client,
        user4_token: auth.AccessToken,
        history_entity_ids: list[str],
    ):
        response = fa_data_client.post(
            f"/history/diff/places/{history_entity_ids[0]}?from_version=1",
            headers=user4_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK
        response_data = response.json()
        diff = response_data["diff"]
        assert "a" == diff[0]["before"]
        assert "aaaaaaaaa" == diff[0]["after"]
        assert response_data["from_version"] == 1
        assert response_data["to_version"] == 9



from typing import List

import injector

from karp.foundation.commands import CommandHandler
from karp.foundation.events import EventHandler
from karp.lex.domain import events as lex_events
from karp.search.domain import commands
from karp.search.domain.errors import IncompleteQuery
from karp.search.application.use_cases import (
    CreateSearchServiceHandler,
    DeletingIndex,
    EntryAddedHandler,
    EntryDeletedHandler,
    EntryUpdatedHandler,
    ReindexingResource,
    ResourcePublishedHandler,
)
from karp.search.application.queries import (
    ResourceViews,
    PreviewEntry,
    PreviewEntryInputDto,
    EntryPreviewDto,
    SearchService,
    QueryRequest,
    StatisticsDto,
)
from karp.search.application.repositories import IndexUnitOfWork, Index, IndexEntry
from karp.search.application.transformers import (
    EntryTransformer,
    PreProcessor,
)


class Search(injector.Module):
    @injector.provider
    def reindex_resource(
        self,
        index_uow: IndexUnitOfWork,
        pre_processor: PreProcessor,
        resource_views: ResourceViews,
    ) -> CommandHandler[commands.ReindexResource]:
        return ReindexingResource(
            index_uow=index_uow,
            pre_processor=pre_processor,
            resource_views=resource_views,
        )

    @injector.multiprovider
    def create_index(
        self, index_uow: IndexUnitOfWork
    ) -> List[EventHandler[lex_events.ResourceCreated]]:
        return [CreateSearchServiceHandler(index_uow)]

    @injector.multiprovider
    def deleting_index(
        self,
        index_uow: IndexUnitOfWork,
    ) -> List[EventHandler[lex_events.ResourceDiscarded]]:
        return [DeletingIndex(index_uow=index_uow)]

    @injector.multiprovider
    def publish_index(
        self, index_uow: IndexUnitOfWork
    ) -> List[EventHandler[lex_events.ResourcePublished]]:
        return [ResourcePublishedHandler(index_uow)]

    @injector.multiprovider
    def add_entry(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ) -> List[EventHandler[lex_events.EntryAdded]]:
        return [
            EntryAddedHandler(
                index_uow=index_uow,
                entry_transformer=entry_transformer,
                resource_views=resource_views,
            )
        ]

    @injector.multiprovider
    def update_entry(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ) -> List[EventHandler[lex_events.EntryUpdated]]:
        return [
            EntryUpdatedHandler(
                index_uow=index_uow,
                entry_transformer=entry_transformer,
                resource_views=resource_views,
            )
        ]

    @injector.multiprovider
    def delete_entry(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ) -> List[EventHandler[lex_events.EntryDeleted]]:
        return [
            EntryDeletedHandler(
                index_uow=index_uow,
                entry_transformer=entry_transformer,
                resource_views=resource_views,
            )
        ]



"""Pytest entry point."""


import pytest
import injector
import pytest  # pyre-ignore
from sqlalchemy import create_engine
from sqlalchemy.orm import session, sessionmaker

from alembic.config import main as alembic_main

from karp.tests.unit.lex.adapters import InMemoryLexInfrastructure
from karp.search_infrastructure import SearchInfrastructure, GenericSearchInfrastructure
from karp.search import Search
from karp.main.modules import CommandBusMod, EventBusMod
from karp.lex_infrastructure import GenericLexInfrastructure
from karp.lex import Lex
from karp.foundation.events import EventBus
from karp.foundation.commands import CommandBus

# environ["TESTING"] = "True"
# environ["ELASTICSEARCH_HOST"] = "localhost:9202"
# environ["CONSOLE_LOG_LEVEL"] = "DEBUG"

from karp.db_infrastructure.db import metadata  # nopep8
from karp.tests.unit.search import adapters as search_adapters
from karp.tests.integration import adapters


@pytest.fixture(name="in_memory_sqlite_db")
def fixture_in_memory_sqlite_db():
    engine = create_engine("sqlite:///:memory:")
    metadata.create_all(engine)
    yield engine
    session.close_all_sessions()
    metadata.drop_all(bind=engine)


@pytest.fixture
def sqlite_session_factory(in_memory_sqlite_db):
    yield sessionmaker(bind=in_memory_sqlite_db)


@pytest.fixture()
def integration_ctx() -> adapters.IntegrationTestContext:
    container = injector.Injector(
        [
            CommandBusMod(),
            EventBusMod(),
            Search(),
            SearchInfrastructure(),
            Lex(),
            GenericLexInfrastructure(),
            InMemoryLexInfrastructure(),
            GenericSearchInfrastructure(),
            search_adapters.InMemorySearchInfrastructure(),
        ],
        auto_bind=False,
    )
    return adapters.IntegrationTestContext(
        container=container,
        command_bus=container.get(CommandBus),  # type: ignore
        event_bus=container.get(EventBus),  # type: ignore
    )



from karp.lex.application.repositories import (
    EntryUowRepositoryUnitOfWork,
)

from . import adapters, factories


class TestCreateEntryRepository:
    def test_create_entry_repository(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd)

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        assert entry_uow_repo_uow.was_committed  # type: ignore [attr-defined]
        assert entry_uow_repo_uow.repo.num_entities() == 1


class TestDeleteEntryRepository:
    def test_delete_entry_repository_succeeds(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd)

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        assert entry_uow_repo_uow.repo.num_entities() == 1

        cmd = factories.DeleteEntryRepositoryFactory(entity_id=cmd.entity_id)
        lex_ctx.command_bus.dispatch(cmd)
        assert entry_uow_repo_uow.repo.num_entities() == 0



import contextlib
from karp.search_infrastructure.repositories.es6_indicies import (
    Es6Index,
    Es6IndexUnitOfWork,
)


class TestEs6Index:
    def test_can_instantiate_es6_index(self):
        with contextlib.suppress(AttributeError):
            Es6Index(es=None, mapping_repo=None)

    def test_can_instantiate_es6_index_uow(self):
        with contextlib.suppress(AttributeError):
            Es6IndexUnitOfWork(
                es=None, event_bus=None, index_prefix="", mapping_repo=None
            )



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

from typing import Any
from dataclasses import dataclass

from tatsu.objectmodel import Node
from tatsu.semantics import ModelBuilderSemantics


class ModelBase(Node):
    pass


class KarpQueryV6ModelBuilderSemantics(ModelBuilderSemantics):
    def __init__(self, context=None, types=None):
        types = [
            t
            for t in globals().values()
            if type(t) is type and issubclass(t, ModelBase)
        ] + (types or [])
        super(KarpQueryV6ModelBuilderSemantics, self).__init__(
            context=context, types=types
        )


@dataclass
class And(ModelBase):
    exps: Any = None
    op: Any = None


@dataclass
class Contains(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Endswith(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Equals(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Exists(ModelBase):
    field: Any = None
    op: Any = None


@dataclass
class Freergxp(ModelBase):
    arg: Any = None
    op: Any = None


@dataclass
class FreetextAnyButString(ModelBase):
    arg: Any = None
    op: Any = None


@dataclass
class FreetextString(ModelBase):
    arg: Any = None
    op: Any = None


@dataclass
class Gt(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Gte(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Lt(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Lte(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Missing(ModelBase):
    field: Any = None
    op: Any = None


@dataclass
class Not(ModelBase):
    exps: Any = None
    op: Any = None


@dataclass
class Or(ModelBase):
    exps: Any = None
    op: Any = None


@dataclass
class Regexp(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None


@dataclass
class Startswith(ModelBase):
    arg: Any = None
    field: Any = None
    op: Any = None



from karp.lex_infrastructure.repositories import SqlEntryUowRepository


def test_create_sql_entry_uow_repo(sqlite_session_factory):
    session = sqlite_session_factory()

    repo = SqlEntryUowRepository(entry_uow_factory=None, session=session)



import os

import environs
from sqlalchemy.engine import URL as DatabaseUrl, make_url
from starlette.config import Config
from starlette.datastructures import Secret

PROJECT_NAME = "Karp"
VERSION = "6.1.5"
API_PREFIX = "/"
# SECRET_KEY = config("SECRET_KEY", cast=Secret, default="CHANGEME")


def load_env() -> environs.Env:
    config_path = os.environ.get("CONFIG_PATH", ".env")
    env = environs.Env()
    env.read_env(config_path)
    return env


def parse_database_name(env: environs.Env) -> str:
    database_name = env("DB_DATABASE", "karp")
    if env("TESTING", None):
        database_name = env("DB_TEST_DATABASE", None) or f"{database_name}_test"
    return database_name


def parse_database_url(env: environs.Env) -> DatabaseUrl:
    database_url = env("DATABASE_URL", None)
    if env.bool("TESTING", False):
        database_test_url = env("DATABASE_TEST_URL", None)
        if database_test_url:
            return make_url(database_test_url)
        elif database_url:
            return make_url(f"{database_url}_test")

    if database_url:
        return make_url(database_url)

    database_name = parse_database_name(env)

    return DatabaseUrl.create(
        drivername=env("DB_DRIVER", "mysql+pymysql"),
        username=env("DB_USER", None),
        password=env("DB_PASSWORD", None),
        host=env("DB_HOST", None),
        port=env.int("DB_PORT", None),
        database=database_name,
        query={"charset": "utf8mb4"},
    )


def parse_database_url_wo_db(env: environs.Env) -> DatabaseUrl:
    return DatabaseUrl.create(
        drivername=env("DB_DRIVER", "mysql+pymysql"),
        username=env("DB_USER", None),
        password=env("DB_PASSWORD", None),
        host=env("DB_HOST", None),
        port=env.int("DB_PORT", None),
        query={"charset": "utf8mb4"},
    )


config = load_env()


DATABASE_URL = parse_database_url(config)
DATABASE_URL_WO_DB = parse_database_url_wo_db(config)
DATABASE_NAME = parse_database_name(config)

AUTH_JWT_AUDIENCE = "spraakbanken:auth"


def parse_sqlalchemy_url_wo_db(env: environs.Env) -> DatabaseUrl:
    return DatabaseUrl.create(
        drivername=env("DB_DRIVER", "mysql+pymysql"),
        username=env("DB_USER", None),
        password=env("DB_PASSWORD", None),
        host=env("DB_HOST", None),
        port=env.int("DB_PORT", None),
        database="",
        query={"charset": "utf8mb4"},
    )



from typing import Callable, Type
from fastapi import Depends
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
from starlette.requests import Request

from karp.db_infrastructure import Database

from karp.lex_infrastructure import SqlResourceUnitOfWork
from karp.lex_infrastructure.repositories import SqlResourceRepository


def get_database(request: Request) -> Database:
    return request.app.state.db


def get_session(request: Request) -> Session:
    return request.state.session


def get_connection(request: Request) -> Connection:
    return request.state.connection



import logging

import typing
from typing import Dict, List, Optional, Tuple, Iterable

from karp.foundation import (
    events as foundation_events,
    commands as foundation_commands,
)
from karp.lex.domain import events as lex_events
from karp.lex.domain.entities import resource
from karp.lex.application.queries import EntryDto
from karp.search.application.queries import ResourceViews
from karp.search.application.repositories import IndexUnitOfWork
from karp.search.application.transformers import EntryTransformer, PreProcessor

from karp.lex.domain import events, entities

# , errors, events, search_service, model
from karp.search.domain import commands

logger = logging.getLogger(__name__)


# def pre_process_resource(
#     resource_obj: Resource,
# ) -> List[Tuple[str, EntryMetadata, Dict]]:
#     metadata = resourcemgr.get_all_metadata(resource_obj)
#     fields = resource_obj.config["fields"].items()
#     entries = resource_obj.entities.query.filter_by(deleted=False)
#     return [
#         (
#             entry.entry_id,
#             metadata[entry.id],
#             transform_to_search_service_entry(resource_obj, json.loads(entry.body), fields),
#         )
#         for entry in entries
#     ]

# metadata = resourcemgr.get_all_metadata(resource_obj)
# fields = resource_obj.config["fields"].items()
# entries = resource_obj.entities.query.filter_by(deleted=False)
# return [
#     (
#         entry.entry_id,
#         metadata[entry.id],
#         transform_to_search_service_entry(resource_obj, json.loads(entry.body), fields),
#     )
#     for entry in entries
# ]


# def research_service(
#     resource_id: str,
#     version: Optional[int] = None,
#     search_entries: Optional[List[Tuple[str, EntryMetadata, Dict]]] = None,
# ) -> None:
#     """
#     If `search_entries` is not given, they will be fetched from DB and processed using `transform_to_search_service_entry`
#     If `search_entries` is given, they most have the same format as the output from `pre_process_resource`
#     """
#     resource_obj = resourcemgr.get_resource(resource_id, version=version)
#     try:
#         search_service_name = search_serviceer.impl.create_search_service(resource_id, resource_obj.config)
#     except NotImplementedError:
#         _logger.error("No SearchService module is loaded. Check your configurations...")
#         sys.exit(errors.NoSearchServiceModuleConfigured)
#     if not search_entries:
#         search_entries = pre_process_resource(resource_obj)
#     add_entries(search_service_name, search_entries, update_refs=False)
#     search_serviceer.impl.publish_search_service(resource_id, search_service_name)
# def research_service(
#     search_serviceer: SearchService,
#     resource_repo: ResourceRepository,
#     resource: Resource,
#     search_entries: Optional[List[IndexEntry]] = None,
# ):


class ReindexingResource(foundation_commands.CommandHandler[commands.ReindexResource]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        resource_views: ResourceViews,
        pre_processor: PreProcessor,
    ) -> None:
        super().__init__()
        self.index_uow = index_uow
        self.resource_views = resource_views
        self.pre_processor = pre_processor

    def execute(self, command: commands.ReindexResource) -> None:
        logger.debug("Reindexing resource '%s'", command.resource_id)
        with self.index_uow as uw:
            uw.repo.create_index(
                command.resource_id,
                self.resource_views.get_resource_config(command.resource_id),
            )
            uw.repo.add_entries(
                command.resource_id, self.pre_processor.process(command.resource_id)
            )
            uw.commit()


class ResourcePublishedHandler(
    foundation_events.EventHandler[lex_events.ResourcePublished]
):
    def __init__(self, index_uow: IndexUnitOfWork):
        self.index_uow = index_uow

    def __call__(
        self,
        event: events.ResourcePublished,
        *args,
        **kwargs,
    ) -> None:
        # research_service(event, ctx)
        with self.index_uow as uw:
            uw.repo.publish_index(event.resource_id)
            uw.commit()
        # if version:
        #     resourcemgr.publish_resource(resource_id, version)


class CreateSearchServiceHandler(
    foundation_events.EventHandler[lex_events.ResourceCreated]
):
    def __init__(self, index_uow: IndexUnitOfWork):
        self.index_uow = index_uow

    def collect_new_events(self) -> Iterable[foundation_events.Event]:
        yield from self.index_uow.collect_new_events()

    def __call__(self, event: events.ResourceCreated, *args, **kwargs):
        with self.index_uow as uw:
            uw.repo.create_index(event.resource_id, event.config)
            uw.commit()


class DeletingIndex(foundation_events.EventHandler[lex_events.ResourceDiscarded]):
    def __init__(self, index_uow: IndexUnitOfWork):
        self.index_uow = index_uow

    def collect_new_events(self) -> Iterable[foundation_events.Event]:
        yield from self.index_uow.collect_new_events()

    def __call__(self, event: events.ResourceDiscarded, *args, **kwargs):
        pass


# def add_entries(
#     resource_id: str,
#     entries: List[Tuple[str, EntryMetadata, Dict]],
#     update_refs: bool = True,
# ) -> None:
#     search_serviceer.impl.add_entries(resource_id, entries)
#     if update_refs:
#         _update_references(resource_id, [entry_id for (entry_id, _, _) in entries])


class EntryAddedHandler(foundation_events.EventHandler[lex_events.EntryAdded]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ):
        self.index_uow = index_uow
        self.entry_transformer = entry_transformer
        self.resource_views = resource_views

    def __call__(
        self,
        event: events.EntryAdded,
        *args,
        **kwargs,
    ):
        with self.index_uow as uw:

            for resource_id in self.resource_views.get_resource_ids(event.repo_id):
                entry = EntryDto(
                    entity_id=event.entity_id,
                    # entry_id=event.entry_id,
                    repository_id=event.repo_id,
                    resource=resource_id,
                    entry=event.body,
                    message=event.message,
                    last_modified=event.timestamp,
                    last_modified_by=event.user,
                    version=1,
                )
                uw.repo.add_entries(
                    resource_id, [self.entry_transformer.transform(resource_id, entry)]
                )
                # self.entry_transformer.update_references(resource_id, [event.entry_id])
            uw.commit()


class EntryUpdatedHandler(foundation_events.EventHandler[lex_events.EntryUpdated]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ):
        self.index_uow = index_uow
        self.entry_transformer = entry_transformer
        self.resource_views = resource_views

    def __call__(
        self,
        event: events.EntryUpdated,
        *args,
        **kwargs,
    ):
        with self.index_uow as uw:

            for resource_id in self.resource_views.get_resource_ids(event.repo_id):
                entry = EntryDto(
                    entity_id=event.entity_id,
                    # entry_id=event.entry_id,
                    repository_id=event.repo_id,
                    resource=resource_id,
                    entry=event.body,
                    message=event.message,
                    last_modified=event.timestamp,
                    last_modified_by=event.user,
                    version=event.version,
                )
                uw.repo.add_entries(
                    resource_id, [self.entry_transformer.transform(resource_id, entry)]
                )
                self.entry_transformer.update_references(resource_id, [event.entity_id])
            uw.commit()
            # add_entries(event.resource_id, [entry], ctx)
            # ctx.index_uow.commit()


class EntryDeletedHandler(foundation_events.EventHandler[lex_events.EntryDeleted]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ):
        self.index_uow = index_uow
        self.entry_transformer = entry_transformer
        self.resource_views = resource_views

    def __call__(self, event: events.EntryDeleted):
        with self.index_uow as uw:
            for resource_id in self.resource_views.get_resource_ids(event.repo_id):
                # uw.repo.delete_entry(resource_id, entry_id=event.entry_id)
                uw.repo.delete_entry(resource_id, entry_id=event.entity_id)
                self.entry_transformer.update_references(resource_id, [event.entity_id])
            uw.commit()



from .basic_ast import Ast
from .parser import Node, is_a, op, parse  # noqa: F401
from tatsu.walkers import NodeWalker
from .karp_query_v6_parser import KarpQueryV6Parser, KarpQueryV6Semantics
from .karp_query_v6_model import KarpQueryV6ModelBuilderSemantics



"""
Query API.

## Query DSL
"""
import logging
from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Path, Query, Security, status

from karp import errors as karp_errors, auth, search
from karp.search.application.queries import SearchService, QueryRequest
from karp.webapp import schemas

from karp.webapp import dependencies as deps
from karp.webapp.dependencies.fastapi_injector import inject_from_req


logger = logging.getLogger(__name__)


router = APIRouter()


@router.get(
    "/entries/{resource_id}/{entry_ids}",
    description="Returns a list of entries matching the given ids",
    name="Get lexical entries by id",
)
def get_entries_by_id(
    resource_id: str = Path(..., description="The resource to perform operation on"),
    entry_ids: str = Path(
        ...,
        description="Comma-separated. The ids to perform operation on.",
        regex=r"^\w(,\w)*",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    logger.debug("webapp.views.get_entries_by_id")
    if not auth_service.authorize(auth.PermissionLevel.read, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    return search_service.search_ids(resource_id, entry_ids)


@router.get("/split/{resources}", name="Query per resource")
def query_split(
    resources: str = Path(
        ...,
        regex=r"^[a-z_0-9\-]+(,[a-z_0-9\-]+)*$",
        description="A comma-separated list of resource identifiers",
    ),
    q: Optional[str] = Query(
        None,
        title="query",
        description="The query. If missing, all entries in chosen resource(s) will be returned.",
    ),
    from_: int = Query(
        0, alias="from", description="Specify which entry should be the first returned."
    ),
    size: int = Query(25, description="Number of entries in page."),
    sort: str = Query(
        None,
        description="The `field` to sort by. If missing, default order for each resource will be used.",
        regex=r"^[a-zA-Z0-9_\-]+(\|asc|desc)?",
    ),
    lexicon_stats: bool = Query(True, description="Show the hit count per lexicon"),
    include_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to return"
    ),
    exclude_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to remove from result"
    ),
    format: schemas.EntryFormat = Query(
        schemas.EntryFormat.json,
        description="Will return the result in the specified format.",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    logger.debug("/query/split called", extra={"resources": resources})
    resource_list = resources.split(",")
    if not auth_service.authorize(auth.PermissionLevel.read, user, resource_list):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    query_request = QueryRequest(
        resource_ids=resource_list,
        q=q,
        from_=from_,
        size=size,
        lexicon_stats=lexicon_stats,
    )
    try:
        response = search_service.query_split(query_request)
    except karp_errors.KarpError as err:
        logger.exception(
            "Error occured when calling '/query/split'",
            extra={"resources": resources, "q": q, "error_message": err.message},
        )
        raise
    except search.IncompleteQuery as err:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail={
                "errorCode": karp_errors.ClientErrorCodes.SEARCH_INCOMPLETE_QUERY,
                "message": "Error in query",
                "failing_query": err.failing_query,
                "error_description": err.error_description,
            },
        )
    return response


@router.get(
    "/{resources}",
    # summary="Returns a list of entries matching the given query in the given resources. The results are mixed from the given resources.",
    name="Query",
    responses={200: {"content": {"application/json": {}}}},
)
def query(
    resources: str = Path(
        ...,
        regex=r"^[a-z_0-9\-]+(,[a-z_0-9\-]+)*$",
        description="A comma-separated list of resource identifiers",
    ),
    q: Optional[str] = Query(
        None,
        title="query",
        description="The query. If missing, all entries in chosen resource(s) will be returned.",
    ),
    from_: int = Query(
        0, alias="from", description="Specify which entry should be the first returned."
    ),
    size: int = Query(25, description="Number of entries in page."),
    sort: List[str] = Query(
        [],
        description="The `field` to sort by. If missing, default order for each resource will be used.",
        regex=r"^[a-zA-Z0-9_\-]+(\|asc|desc)?",
    ),
    lexicon_stats: bool = Query(True, description="Show the hit count per lexicon"),
    include_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to return"
    ),
    exclude_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to remove from result"
    ),
    format_: schemas.EntryFormat = Query(
        schemas.EntryFormat.json,
        alias="format",
        description="Will return the result in the specified format.",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    """
    Returns a list of entries matching the given query in the given resources. The results are mixed from the given resources.

    """
    logger.debug(
        "Called 'query' called with",
        extra={"resources": resources, "from": from_, "size": size},
    )
    resource_list = resources.split(",")
    if not auth_service.authorize(auth.PermissionLevel.read, user, resource_list):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    query_request = QueryRequest(
        resource_ids=resource_list,
        q=q,
        from_=from_,
        size=size,
        sort=sort,
        include_fields=include_fields,
        exclude_fields=exclude_fields,
        format_=format_,
        lexicon_stats=lexicon_stats,
    )
    try:
        print(f"{search_service=}")
        response = search_service.query(query_request)

    except karp_errors.KarpError as err:
        logger.exception(
            "Error occured when calling 'query' with",
            extra={"resources": resources, "q": q, "error_message": err.message},
        )
        raise
    return response


def init_app(app):
    app.include_router(router)



from typing import Callable

import pytest

from karp.lex.domain.events import ResourceCreated

from karp.search.application.repositories import IndexUnitOfWork
from karp.tests.unit.lex import factories as lex_factories
from . import adapters, factories


@pytest.fixture(name="resource_created")
def fixture_resource_created() -> ResourceCreated:
    return lex_factories.ResourceCreatedFactory()  # type: ignore [return-value]


@pytest.mark.parametrize(
    "event_factory,predicate",
    [
        (None, lambda x: x.created),
        (lex_factories.ResourcePublishedFactory, lambda x: x.published),
    ],
)
def test_index_reacts_on_lex_events(
    event_factory,
    predicate: Callable,
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    resource_created = lex_factories.ResourceCreatedFactory()
    search_unit_ctx.event_bus.post(resource_created)  # type: ignore [arg-type]
    if event_factory:
        event = event_factory(resource_id=resource_created.resource_id)
        search_unit_ctx.event_bus.post(event)

    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)  # type: ignore [misc]
    assert index_uow.was_committed  # type: ignore [attr-defined]

    assert predicate(index_uow.repo.indicies[resource_created.resource_id])  # type: ignore [attr-defined]


def test_index_reacts_on_entry_added_event(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)  # type: ignore [arg-type]
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id
    )
    search_unit_ctx.command_bus.dispatch(create_resource)  # type: ignore [arg-type]
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={"baseform": "bra"},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)  # type: ignore [arg-type]
    # if event_factory:
    #     event = event_factory(resource_id=create_resource.resource_id)
    #     search_unit_ctx.event_bus.post(event)

    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)  # type: ignore [misc]
    assert index_uow.was_committed  # type: ignore [attr-defined]

    resource_index = index_uow.repo.indicies[create_resource.resource_id]  # type: ignore [attr-defined]
    assert resource_index.created
    assert len(resource_index.entries) == 1
    assert str(create_entry.entity_id) in resource_index.entries


def test_index_reacts_on_entry_updated_event(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)  # type: ignore [arg-type]
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id
    )
    search_unit_ctx.command_bus.dispatch(create_resource)  # type: ignore [arg-type]
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={"baseform": "bra"},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)  # type: ignore [arg-type]

    update_entry = lex_factories.UpdateEntryFactory(
        resource_id=create_resource.resource_id,
        entity_id=create_entry.entity_id,
        # entry_id="bra",
        entry={"baseform": "bra", "wordclass": "adjektiv"},
        version=1,
    )
    search_unit_ctx.command_bus.dispatch(update_entry)  # type: ignore [arg-type]

    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)  # type: ignore [misc]
    assert index_uow.was_committed  # type: ignore [attr-defined]
    resource_index = index_uow.repo.indicies[create_resource.resource_id]  # type: ignore [attr-defined]
    assert resource_index.created
    assert len(resource_index.entries) == 1
    entry = resource_index.entries[str(update_entry.entity_id)]
    assert entry.entry["wordclass"] == "adjektiv"


def test_index_reacts_on_entry_deleted_event(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)  # type: ignore [arg-type]
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id
    )
    search_unit_ctx.command_bus.dispatch(create_resource)  # type: ignore [arg-type]
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={"baseform": "bra"},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)  # type: ignore [arg-type]

    delete_entry = lex_factories.DeleteEntryFactory(
        entity_id=create_entry.entity_id,
        resource_id=create_resource.resource_id,
        entry_id="bra",
        version=1,
    )
    search_unit_ctx.command_bus.dispatch(delete_entry)  # type: ignore [arg-type]

    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)  # type: ignore [misc]
    assert index_uow.was_committed  # type: ignore [attr-defined]

    resource_index = index_uow.repo.indicies[create_resource.resource_id]  # type: ignore [attr-defined]
    assert resource_index.created
    assert len(resource_index.entries) == 0
    assert "bra" not in resource_index.entries


def test_reindex_resource_command(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)  # type: ignore [arg-type]
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id
    )
    search_unit_ctx.command_bus.dispatch(create_resource)  # type: ignore [arg-type]
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={"baseform": "bra"},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)  # type: ignore [arg-type]

    reindex_resource = factories.ReindexResourceFactory(
        resource_id=create_resource.resource_id,
        version=1,
    )
    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)  # type: ignore [misc]
    resource_index = index_uow.repo.indicies[create_resource.resource_id]  # type: ignore [attr-defined]
    previous_created_at = resource_index.created_at

    search_unit_ctx.command_bus.dispatch(reindex_resource)  # type: ignore [arg-type]

    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)  # type: ignore [misc]
    assert index_uow.was_committed  # type: ignore [attr-defined]

    assert resource_index.created
    assert len(resource_index.entries) == 1
    assert resource_index.created_at >= previous_created_at



from typing import Dict, Optional

import pytest

from karp.search_infrastructure.repositories.es6_indicies import create_es6_mapping


class TestCreateEs6Mapping:
    def test_minimal_valid_input(self):
        data = {"fields": {}}
        mapping = create_es6_mapping(data)

        assert mapping["properties"] == {}

    @pytest.mark.parametrize(
        "field, field_def, expected_property",
        [
            ("name", {"type": "boolean"}, {}),
            ("name", {"type": "number"}, {"type": "double"}),
            ("name", {"type": "integer"}, {"type": "long"}),
            (
                "name",
                {"type": "string"},
                {"type": "text", "fields": {"raw": {"type": "keyword"}}},
            ),
            ("name", {"type": "long_string"}, {"type": "text"}),
        ],
    )
    def test_standard_types(
        self, field: str, field_def: Dict, expected_property: Optional[Dict]
    ):
        data = {
            "fields": {
                field: field_def,
            },
        }
        mapping = create_es6_mapping(data)
        assert field in mapping["properties"]
        expected_property = expected_property or field_def
        assert mapping["properties"][field] == expected_property

    @pytest.mark.parametrize(
        "field, field_def, expected_property",
        [
            (
                "name",
                {"type": "object", "fields": {"first": {"type": "string"}}},
                {
                    "properties": {
                        "first": {
                            "type": "text",
                            "fields": {"raw": {"type": "keyword"}},
                        }
                    }
                },
            ),
            ("name", {"type": "number"}, {"type": "double"}),
            ("name", {"type": "integer"}, {"type": "long"}),
            (
                "name",
                {"type": "string"},
                {"type": "text", "fields": {"raw": {"type": "keyword"}}},
            ),
            ("name", {"type": "long_string"}, {"type": "text"}),
        ],
    )
    def test_complex_types(
        self, field: str, field_def: Dict, expected_property: Optional[Dict]
    ):
        data = {
            "fields": {
                field: field_def,
            },
        }
        mapping = create_es6_mapping(data)
        assert field in mapping["properties"]
        expected_property = expected_property or field_def
        assert mapping["properties"][field] == expected_property

    def test_sort(self):
        data = {"fields": {"name": {"type": "string"}}, "sort": ["name"]}

        mapping = create_es6_mapping(data)

        expected = {
            "dynamic": False,
            "properties": {
                "name": {"type": "text", "fields": {"raw": {"type": "keyword"}}}
            },
            "settings": {
                "analysis": {
                    "analyzer": {
                        "default": {
                            "char_filter": [
                                "compound",
                                "swedish_aa",
                                "swedish_ae",
                                "swedish_oe",
                            ],
                            "filter": ["swedish_folding", "lowercase"],
                            "tokenizer": "standard",
                        },
                    },
                    "char_filter": {
                        "compound": {
                            "pattern": "-",
                            "replacement": "",
                            "type": "pattern_replace",
                        },
                        "swedish_aa": {
                            "pattern": "[Ǻǻ]",
                            "replacement": "å",
                            "type": "pattern_replace",
                        },
                        "swedish_ae": {
                            "pattern": "[æÆǞǟ]",
                            "replacement": "ä",
                            "type": "pattern_replace",
                        },
                        "swedish_oe": {
                            "pattern": "[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]",
                            "replacement": "ö",
                            "type": "pattern_replace",
                        },
                    },
                    "filter": {
                        "swedish_folding": {
                            "type": "icu_folding",
                            "unicodeSetFilter": "[^åäöÅÄÖ]",
                        },
                        "swedish_sort": {"language": "sv", "type": "icu_collation"},
                    },
                }
            },
        }

        assert mapping == expected

    @pytest.mark.skip()
    def test_sort_large(self):
        data = {"fields": {"name": {"type": "string"}}, "sort": ["name"]}

        mapping = create_es6_mapping(data)

        expected = {
            "properties": {
                "name": {"type": "text", "fields": {"raw": {"type": "keyword"}}}
            },
            "settings": {
                "analysis": {
                    "analyzer": {
                        "default": {
                            "char_filter": [
                                "compound",
                                "swedish_aa",
                                "swedish_ae",
                                "swedish_oe",
                            ],
                            "filter": ["swedish_folding", "lowercase"],
                            "tokenizer": "standard",
                        },
                    },
                    "char_filter": {
                        "clean_lastname": {
                            "pattern": "^(af|von) ",
                            "replacement": "",
                            "type": "pattern_replace",
                        },
                        "compound": {
                            "pattern": "-",
                            "replacement": "",
                            "type": "pattern_replace",
                        },
                        "init": {
                            "pattern": "^(.).*",
                            "replacement": "$1",
                            "type": "pattern_replace",
                        },
                        "links": {
                            "pattern": "\\[([^\\]]*)\\](\\([^\\)]*\\))?",
                            "replacement": "$1",
                            "type": "pattern_replace",
                        },
                        "pics": {
                            "pattern": "!\\[[^\\]]*\\](\\([^\\)]*\\))?",
                            "replacement": "",
                            "type": "pattern_replace",
                        },
                        "remove_slash_filter": {
                            "pattern": "(.*)/(.+)/(.*)",
                            "replacement": "$1$2$3",
                            "type": "pattern_replace",
                        },
                        "removespecials": {
                            "pattern": "[^A-Za-zÀÁÂÄÅÆÇÈÉÊËÌÍÎÏÑÒÓÔÖØÙÚÛÜàáâäåæçèéêëìíîïñòóôøöùúûü]",
                            "replacement": "",
                            "type": "pattern_replace",
                        },
                        "slash_filter": {
                            "pattern": "(.*)/(.+)/(.*)",
                            "replacement": "$2",
                            "type": "pattern_replace",
                        },
                        "swedish_aa": {
                            "pattern": "[Ǻǻ]",
                            "replacement": "å",
                            "type": "pattern_replace",
                        },
                        "swedish_ae": {
                            "pattern": "[æÆǞǟ]",
                            "replacement": "ä",
                            "type": "pattern_replace",
                        },
                        "swedish_oe": {
                            "pattern": "[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]",
                            "replacement": "ö",
                            "type": "pattern_replace",
                        },
                        "underscore": {
                            "pattern": "_*((~_)*)_*",
                            "replacement": "$1",
                            "type": "pattern_replace",
                        },
                    },
                    "filter": {
                        "swedish_folding": {
                            "type": "icu_folding",
                            "unicodeSetFilter": "[^åäöÅÄÖ]",
                        },
                        "swedish_sort": {"language": "sv", "type": "icu_collation"},
                    },
                }
            },
        }

        assert mapping == expected



from .machine_name import MachineName
from .permission_level import PermissionLevel
from .unique_id import UniqueId, make_unique_id, UniqueIdStr

__all__ = ["PermissionLevel", "UniqueId", "UniqueIdStr", "make_unique_id"]



"""SQL repositories for entries."""
import logging
import typing
from typing import Dict, List, Optional, Generic, TypeVar

import injector
import regex
import sqlalchemy as sa
from sqlalchemy import func as sa_func
from sqlalchemy import sql
from sqlalchemy.orm import sessionmaker
import ulid

from karp.foundation.value_objects import UniqueId
from karp.foundation.events import EventBus
from karp.lex.domain import errors
from karp.lex.application import repositories

# from karp.domain.errors import NonExistingField, RepositoryError
from karp.lex.domain.entities.entry import (  # EntryRepositorySettings,; EntryRepository,; create_entry_repository,
    Entry,
    EntryOp,
    EntryStatus,
)
from karp.db_infrastructure import db
from karp.lex_infrastructure.sql import sql_models
from karp.db_infrastructure.sql_repository import SqlRepository
from karp.db_infrastructure.sql_unit_of_work import SqlUnitOfWork

logger = logging.getLogger(__name__)

DUPLICATE_PATTERN = r"Duplicate entry '(.+)' for key '(\w+)'"
DUPLICATE_PROG = regex.compile(DUPLICATE_PATTERN)
NO_PROPERTY_PATTERN = regex.compile(r"has no property '(\w+)'")


class SqlEntryRepository(repositories.EntryRepository, SqlRepository):
    def __init__(
        self,
        history_model,
        resource_config: Dict,
        *,
        session: db.Session,
    ):
        if not session:
            raise TypeError("session can't be None")
        repositories.EntryRepository.__init__(self)
        SqlRepository.__init__(self, session=session)
        self.history_model = history_model
        self.resource_config = resource_config

    @classmethod
    def from_dict(
        cls,
        name: str,
        resource_config: typing.Dict,
        *,
        session: db.Session,
    ):
        if not session:
            raise TypeError("session can't be None")
        table_name = name

        logger.info({"table_name": table_name})
        history_model = sql_models.get_or_create_entry_history_model(table_name)

        if session:
            history_model.__table__.create(  # type:ignore [attr-defined]
                bind=session.bind, checkfirst=True
            )

        return cls(
            history_model=history_model,
            resource_config=resource_config,
            session=session,
        )

    @classmethod
    def _create_repository_settings(
        cls, resource_id: str, resource_config: typing.Dict
    ) -> typing.Dict:
        return {
            "table_name": resource_id,
            "resource_id": resource_id,
        }

    def _save(self, entry: Entry):
        self._check_has_session()

        return self._insert_history(entry)

    def _insert_history(self, entry: Entry):
        self._check_has_session()
        try:
            ins_stmt = db.insert(self.history_model)
            history_dict = self._entry_to_history_dict(entry)
            ins_stmt = ins_stmt.values(**history_dict)
            result = self._session.execute(ins_stmt)
            return result.lastrowid or result.returned_defaults["history_id"]
        except db.exc.DBAPIError as exc:
            raise errors.RepositoryError("db failure") from exc

    def entity_ids(self) -> List[str]:
        stmt = self._stmt_latest_not_discarded()
        stmt = stmt.order_by(self.history_model.last_modified.desc())
        query = self._session.execute(stmt).scalars()
        # query = self._session.query(self.history_model).filter_by(discarded=False)
        return [row.entity_id for row in query.all()]

    def _by_id(
        self,
        id: str,
        *,
        version: Optional[int] = None,
        after_date: Optional[float] = None,
        before_date: Optional[float] = None,
        oldest_first: bool = False,
    ) -> Optional[Entry]:
        self._check_has_session()
        query = self._session.query(self.history_model)
        query = query.filter_by(entity_id=id)
        if version:
            query = query.filter_by(version=version)
        elif after_date is not None:
            query = query.filter(
                self.history_model.last_modified >= after_date
            ).order_by(self.history_model.last_modified)
        elif before_date:
            query = query.filter(
                self.history_model.last_modified <= before_date
            ).order_by(self.history_model.last_modified.desc())
        elif oldest_first:
            query = query.order_by(self.history_model.last_modified)
        else:
            query = query.order_by(self.history_model.last_modified.desc())
        row = query.first()
        return self._history_row_to_entry(row) if row else None

    def teardown(self):
        """Use for testing purpose."""
        logger.info("starting teardown")

        logger.info("droping history_model ...")
        self.history_model.__table__.drop(bind=self._session.bind)
        logger.info("dropped history_model")

    def all_entries(self) -> typing.Iterable[Entry]:
        stmt = self._stmt_latest_not_discarded()
        query = self._session.execute(stmt).scalars()

        return [self._history_row_to_entry(db_entry) for db_entry in query.all()]

    # TODO Rename this here and in `entity_ids` and `all_entries`
    def _stmt_latest_not_discarded(self):
        self._check_has_session()
        subq = self._subq_for_latest()
        return sql.select(self.history_model).join(
            subq,
            sa.and_(
                self.history_model.entity_id == subq.c.entity_id,
                self.history_model.last_modified == subq.c.maxdate,
                self.history_model.discarded == False,
            ),
        )

    def _subq_for_latest(self) -> sql.Subquery:
        return (
            sql.select(
                self.history_model.entity_id,
                sa.func.max(self.history_model.last_modified).label("maxdate"),
            )
            .group_by(self.history_model.entity_id)
            .subquery("t2")
        )

    def num_entities(self) -> int:
        self._check_has_session()

        subq = self._subq_for_latest()

        stmt = sql.select(sa_func.count(self.history_model.entity_id)).join(
            subq,
            sa.and_(
                self.history_model.entity_id == subq.c.entity_id,
                self.history_model.last_modified == subq.c.maxdate,
                self.history_model.discarded == False,
            ),
        )
        return self._session.execute(stmt).scalar()

    def by_referenceable(self, filters: Optional[Dict] = None, **kwargs) -> List[Entry]:
        self._check_has_session()

        if filters is None:
            if not kwargs:
                raise ValueError("Must give 'filters' or kwargs")
            else:
                filters = kwargs

        subq = self._subq_for_latest()
        query_and = sa.and_(
            self.history_model.entity_id == subq.c.entity_id,
            self.history_model.last_modified == subq.c.maxdate,
            self.history_model.discarded == False,
        )

        for filter_key, filter_value in filters.items():

            query_and = query_and & (
                sa_func.json_extract(self.history_model.body, f"$.{filter_key}")
                == filter_value
            )
        stmt = sql.select(self.history_model).join(
            subq,
            query_and,
        )

        return [
            self._history_row_to_entry(row)
            for row in self._session.execute(stmt).scalars().all()
        ]

    def get_history(
        self,
        user_id: Optional[str] = None,
        entry_id: Optional[str] = None,
        from_date: Optional[float] = None,
        to_date: Optional[float] = None,
        from_version: Optional[int] = None,
        to_version: Optional[int] = None,
        offset: int = 0,
        limit: int = 100,
    ):
        self._check_has_session()
        query = self._session.query(self.history_model)
        if user_id:
            query = query.filter_by(last_modified_by=user_id)
        if entry_id:
            query = query.filter_by(entity_id=entry_id)
        if entry_id and from_version:
            query = query.filter(self.history_model.version >= from_version)
        elif from_date is not None:
            query = query.filter(self.history_model.last_modified >= from_date)
        if entry_id and to_version:
            query = query.filter(self.history_model.version < to_version)
        elif to_date is not None:
            query = query.filter(self.history_model.last_modified <= to_date)

        paged_query = query.limit(limit).offset(offset)
        total = query.count()
        return [self._history_row_to_entry(row) for row in paged_query.all()], total

    def _entry_to_history_dict(
        self, entry: Entry, history_id: Optional[int] = None
    ) -> Dict:
        return {
            "history_id": history_id,
            "entity_id": entry.entity_id,
            "version": entry.version,
            "last_modified": entry.last_modified,
            "last_modified_by": entry.last_modified_by,
            "body": entry.body,
            "status": entry.status,
            "message": entry.message,
            "op": entry.op,
            "discarded": entry.discarded,
            "repo_id": entry.repo_id,
        }

    def _history_row_to_entry(self, row) -> Entry:
        return Entry(
            body=row.body,
            message=row.message,
            status=row.status,
            op=row.op,
            entity_id=row.entity_id,
            last_modified=row.last_modified,
            last_modified_by=row.last_modified_by,
            discarded=row.discarded,
            version=row.version,
            repository_id=row.repo_id,
        )


class SqlEntryUnitOfWork(
    SqlUnitOfWork,
    repositories.EntryUnitOfWork,
):
    repository_type: str = "sql_entries_base"

    def __init__(
        self,
        session_factory: sessionmaker,
        event_bus: EventBus,
        **kwargs,
    ):
        SqlUnitOfWork.__init__(self)
        repositories.EntryUnitOfWork.__init__(self, event_bus=event_bus, **kwargs)
        self.session_factory = session_factory
        self._entries = None
        self._session = None

    def _begin(self):
        if self._session is None:
            self._session = self.session_factory()
        if self._entries is None:
            self._entries = SqlEntryRepository.from_dict(
                name=self.table_name(),
                resource_config=self.config,
                session=self._session,
            )
        return self

    def table_name(self) -> str:
        return self.name

    @property
    def repo(self) -> SqlEntryRepository:
        if self._entries is None:
            raise RuntimeError("No entries")
        return self._entries

    @classmethod
    def from_dict(cls, settings: typing.Dict, resource_config, **kwargs):
        return cls(repo_settings=settings, resource_config=resource_config, **kwargs)

    def collect_new_events(self) -> typing.Iterable:
        return super().collect_new_events() if self._entries else []


class SqlEntryUnitOfWorkV1(SqlEntryUnitOfWork):
    repository_type: str = "sql_entries_v1"


class SqlEntryUnitOfWorkV2(SqlEntryUnitOfWork):
    repository_type: str = "sql_entries_v2"

    def table_name(self) -> str:
        u = ulid.from_uuid(self.entity_id)
        random_part = u.randomness().str
        return f"{self.name}_{random_part}"


SqlEntryUowType = TypeVar("SqlEntryUowType", bound=SqlEntryUnitOfWork)


class SqlEntryUowCreator(Generic[SqlEntryUowType]):
    repository_type: str = "repository_type"

    @injector.inject
    def __init__(
        self,
        session_factory: sessionmaker,
        event_bus: EventBus,
    ):
        self._session_factory = session_factory
        self.event_bus = event_bus
        self.cache = {}

    def _create_uow(self, **kwargs) -> SqlEntryUowType:
        raise NotImplementedError(f"please implement this for {self}")

    def __call__(
        self,
        entity_id: UniqueId,
        name: str,
        config: Dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> SqlEntryUowType:
        if entity_id not in self.cache:
            self.cache[entity_id] = self._create_uow(
                entity_id=entity_id,
                name=name,
                config=config,
                connection_str=connection_str,
                last_modified_by=user,
                message=message,
                last_modified=timestamp,
                session_factory=self._session_factory,
                event_bus=self.event_bus,
            )
        return self.cache[entity_id]


class SqlEntryUowV1Creator(SqlEntryUowCreator[SqlEntryUnitOfWorkV1]):
    repository_type: str = "sql_entries_v1"

    def _create_uow(self, **kwargs) -> SqlEntryUnitOfWorkV1:
        return SqlEntryUnitOfWorkV1(**kwargs)


class SqlEntryUowV2Creator(SqlEntryUowCreator[SqlEntryUnitOfWorkV2]):
    repository_type: str = "sql_entries_v2"

    def _create_uow(self, **kwargs) -> SqlEntryUnitOfWorkV2:
        return SqlEntryUnitOfWorkV2(**kwargs)



import copy

import pytest

from karp.lex.domain import errors

from karp.lex.application import repositories
from karp.lex.application.repositories import ResourceUnitOfWork
from karp.lex.domain import commands

from . import adapters, factories

# type: ignore [arg-type]


class TestCreateResource:
    def test_create_resource_w_no_entry_repo_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd = factories.CreateResourceFactory()
        with pytest.raises(errors.NoSuchEntryRepository):
            lex_ctx.command_bus.dispatch(cmd)

    def test_create_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd)

        cmd = factories.CreateResourceFactory(entry_repo_id=cmd.entity_id)
        lex_ctx.command_bus.dispatch(cmd)

        resource_uow = lex_ctx.container.get(ResourceUnitOfWork)  # type: ignore [misc]
        assert resource_uow.was_committed  # type: ignore [attr-defined]
        assert resource_uow.resources.num_entities() == 1

        resource = resource_uow.resources.by_id(cmd.entity_id)
        assert resource.entity_id == cmd.entity_id
        assert resource.resource_id == cmd.resource_id
        # assert len(resource.domain_events) == 1

    def test_create_resource_with_same_resource_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)
        cmd2 = factories.CreateResourceFactory(entry_repo_id=cmd1.entity_id)
        lex_ctx.command_bus.dispatch(cmd2)

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.CreateResourceFactory(
                    resource_id=cmd2.resource_id, entry_repo_id=cmd1.entity_id
                )
            )

        resource_uow = lex_ctx.container.get(repositories.ResourceUnitOfWork)  # type: ignore [misc]

        assert resource_uow.was_rolled_back  # type: ignore [attr-defined]

    def test_bad_resource_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd)
        with pytest.raises(errors.InvalidResourceId):
            lex_ctx.command_bus.dispatch(
                factories.CreateResourceFactory(
                    entry_repo_id=cmd.entity_id,
                    resource_id="with space",
                )
            )


class TestUpdateResource:
    def test_update_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)
        cmd2 = factories.CreateResourceFactory(entry_repo_id=cmd1.entity_id)
        lex_ctx.command_bus.dispatch(cmd2)

        changed_config = copy.deepcopy(cmd2.config)
        changed_config["fields"]["b"] = {"type": "integer"}
        lex_ctx.command_bus.dispatch(
            factories.UpdateResourceFactory(
                resource_id=cmd2.resource_id,
                version=1,
                name="R1",
                config=changed_config,
                message="changed",
                user="bob",
            ),
        )

        resource_uow = lex_ctx.container.get(ResourceUnitOfWork)  # type: ignore [misc]
        assert resource_uow.was_committed  # type: ignore

        resource = resource_uow.resources.by_resource_id(cmd2.resource_id)
        assert resource is not None
        assert resource.config == changed_config
        assert resource.version == 2

        resource = resource_uow.resources.get_by_id(cmd2.entity_id)
        assert resource is not None
        assert resource.config == changed_config
        assert resource.version == 2


class TestPublishResource:
    def test_publish_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]
        cmd2 = factories.CreateResourceFactory(entry_repo_id=cmd1.entity_id)
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        lex_ctx.command_bus.dispatch(
            commands.PublishResource(
                resource_id=cmd2.resource_id,
                message="publish",
                user="kristoff@example.com",
                version=1,
            )
        )

        resource_uow = lex_ctx.container.get(ResourceUnitOfWork)  # type: ignore [misc]
        assert resource_uow.was_committed  # type: ignore [attr-defined]

        resource = resource_uow.resources.by_id(cmd2.entity_id)  # type: ignore [arg-type]
        assert resource.is_published
        assert resource.version == 2
        # assert index_uow.repo.indicies[resource_id].published
        # assert index_uow.was_committed #type: ignore [attr-defined]



from typing import Optional, Union

from .basic_ast import Ast
from .errors import ParseError, SyntaxError
from .node import Node, create_binary_node, create_unary_node
from .token import Token


class op:
    AND = "AND"  # noqa: E221
    NOT = "NOT"  # noqa: E221
    OR = "OR"  # noqa: E221
    AND_OR = [AND, OR]  # noqa: E221
    LOGICAL = [AND, OR, NOT]  # noqa: E221
    ARG_AND = "ARG_AND"  # noqa: E221
    ARG_OR = "ARG_OR"  # noqa: E221
    ARG_NOT = "ARG_NOT"  # noqa: E221
    ARG_LOGICAL = [ARG_AND, ARG_OR, ARG_NOT]  # noqa: E221
    FREETEXT = "FREETEXT"  # noqa: E221
    FREERGXP = "FREERGXP"  # noqa: E221
    EXISTS = "EXISTS"  # noqa: E221
    MISSING = "MISSING"  # noqa: E221
    UNARY_OPS = [FREETEXT, FREERGXP, EXISTS, MISSING]  # noqa: E221
    EQUALS = "EQUALS"  # noqa: E221
    GT = "GT"  # noqa: E221
    GTE = "GTE"  # noqa: E221
    LT = "LT"  # noqa: E221
    LTE = "LTE"  # noqa: E221
    RANGE_OPS = [GT, GTE, LT, LTE]  # noqa: E221
    CONTAINS = "CONTAINS"  # noqa: E221
    STARTSWITH = "STARTSWITH"  # noqa: E221
    ENDSWITH = "ENDSWITH"  # noqa: E221
    REGEXP = "REGEXP"  # noqa: E221
    BINARY_OPS = [
        EQUALS,
        GT,
        GTE,
        LT,
        LTE,
        CONTAINS,
        STARTSWITH,
        ENDSWITH,
        REGEXP,
    ]  # noqa: E221
    REGEX_OPS = [CONTAINS, STARTSWITH, ENDSWITH, REGEXP]  # noqa: E221
    OPS = [  # noqa: E221
        CONTAINS,
        ENDSWITH,
        EQUALS,
        EXISTS,
        FREERGXP,
        FREETEXT,
        GT,
        GTE,
        LT,
        LTE,
        MISSING,
        REGEXP,
        STARTSWITH,
    ]
    INT = "INT"  # noqa: E221
    FLOAT = "FLOAT"  # noqa: E221
    STRING = "STRING"  # noqa: E221
    ARGS = [INT, FLOAT, STRING]  # noqa: E221
    SEP = "||"  # noqa: E221


def is_a(x: Union[Node, Token], type_) -> bool:
    if isinstance(type_, list):
        return x.type in type_
    else:
        return x.type == type_


def arg_token_any(s) -> Token:
    try:
        v = int(s)
        return Token(op.INT, v)
    except ValueError:
        pass
    try:
        v = float(s)
        return Token(op.FLOAT, v)
    except ValueError:
        pass

    return Token(op.STRING, s)


def arg_token_string(s) -> Token:
    return Token(op.STRING, s)


class KarpTNGLexer:
    SEPARATOR_1 = "||"  # noqa: E221
    SEPARATOR_2 = "|"  # noqa: E221
    logical = {
        "and": op.AND,
        "not": op.NOT,
        "or": op.OR,
    }
    arg_logical = {
        "and": op.ARG_AND,
        "or": op.ARG_OR,
        "not": op.ARG_NOT,
    }
    ops = {
        "freetext": op.FREETEXT,
        "freergxp": op.FREERGXP,
        "equals": op.EQUALS,
        "exists": op.EXISTS,
        "missing": op.MISSING,
        "contains": op.CONTAINS,
        "startswith": op.STARTSWITH,
        "endswith": op.ENDSWITH,
        "regexp": op.REGEXP,
        "gt": op.GT,
        "gte": op.GTE,
        "lt": op.LT,
        "lte": op.LTE,
    }
    arg1 = {
        "freetext": arg_token_any,
        "freergxp": arg_token_string,
        "equals": arg_token_string,
        "exists": arg_token_string,
        "missing": arg_token_string,
        "contains": arg_token_string,
        "startswith": arg_token_string,
        "endswith": arg_token_string,
        "regexp": arg_token_string,
        "gt": arg_token_string,
        "gte": arg_token_string,
        "lt": arg_token_string,
        "lte": arg_token_string,
    }
    arg2 = {
        "equals": arg_token_any,
        "contains": arg_token_string,
        "startswith": arg_token_string,
        "endswith": arg_token_string,
        "regexp": arg_token_string,
        "gt": arg_token_any,
        "gte": arg_token_any,
        "lt": arg_token_any,
        "lte": arg_token_any,
    }

    def tokenize(self, s: str):
        print("Tokenizing {s}".format(s=s))
        exprs = s.split(self.SEPARATOR_1)
        arg_types = []
        for expr in exprs:
            logical_type = self.logical.get(expr)
            if logical_type:
                yield Token(logical_type)
            else:
                print("Tokenizing {expr}".format(expr=expr))
                sub_exprs = expr.split(self.SEPARATOR_2)

                if sub_exprs[0] in self.ops:
                    yield Token(self.ops[sub_exprs[0]])
                    arg_1 = self.arg1[sub_exprs[0]]
                    arg_2 = self.arg2.get(sub_exprs[0])
                    if len(sub_exprs) > 1:
                        yield arg_1(sub_exprs[1])
                        arg_1 = None
                        if len(sub_exprs) > 2:
                            if arg_2:
                                yield arg_2(sub_exprs[2])
                                arg_2 = None
                            else:
                                raise SyntaxError(
                                    "Too many arguments to '{op}' in '{expr}'".format(
                                        op=sub_exprs[0], expr=expr
                                    )
                                )
                    if arg_2:
                        arg_types.append(arg_2)
                    if arg_1:
                        arg_types.append(arg_1)
                else:
                    if sub_exprs[0] in self.arg_logical:
                        yield Token(self.arg_logical[sub_exprs[0]])
                        arg_exprs = sub_exprs[1:]
                    else:
                        arg_exprs = sub_exprs
                    arg = arg_types.pop()
                    if not arg:
                        raise ParseError("No arg type is set")
                    for arg_expr in arg_exprs:
                        yield arg(arg_expr)

            yield Token(op.SEP)


def create_node(tok: Token):
    if is_a(tok, op.UNARY_OPS) or is_a(tok, op.NOT):
        return create_unary_node(tok.type, tok.value)
    elif is_a(tok, op.BINARY_OPS):
        return create_binary_node(tok.type, tok.value)
    elif is_a(tok, op.ARGS):
        return Node(tok.type, 0, tok.value)
    elif is_a(tok, op.LOGICAL) or is_a(tok, op.ARG_LOGICAL):
        return Node(tok.type, None, tok.value)
    else:
        return None


class KarpTNGParser:
    def parse(self, tokens):
        curr = None
        stack = []
        for tok in tokens:
            print("Token({}, {})".format(tok.type, tok.value))
            if is_a(tok, op.ARGS):
                n = create_node(tok)
                if curr:
                    curr.add_child(n)
                elif stack:
                    stack[-1].add_child(n)
            elif is_a(tok, op.SEP):
                if curr:
                    if is_a(curr, op.OPS) and curr.n_children() < curr.arity:
                        stack.append(curr)
                    elif is_a(curr, op.ARG_LOGICAL):
                        n = stack.pop()
                        if is_a(n, op.OPS):
                            n.add_child(curr)
                            stack.append(n)
                        else:
                            raise ParseError(
                                "No OP to add ARG_LOGICAL '{curr}'".format(curr=curr)
                            )
                    elif stack:
                        n1 = stack.pop()
                        n1.add_child(curr)
                        if is_a(n1, op.NOT):
                            if stack:
                                n2 = stack.pop()
                                n2.add_child(n1)
                                stack.append(n2)
                            else:
                                stack.append(n1)
                        else:
                            stack.append(n1)
                    else:
                        stack.append(curr)
                    curr = None
            elif is_a(tok, op.LOGICAL):
                stack.append(create_node(tok))
            else:
                if curr:
                    raise RuntimeError("")
                curr = create_node(tok)
            print("curr = {curr}".format(curr=curr))
            print("stack = {stack}".format(stack=stack))

        root = None
        for node in reversed(stack):
            if root:
                # if is_a(root, op.AND_OR) and is_a(node, op.AND_OR):

                node.add_child(root)
            root = node
        return root


_lexer = KarpTNGLexer()
_parser = KarpTNGParser()


def parse(s: Optional[str]) -> Ast:
    if not s:
        return Ast()
    return Ast(_parser.parse(_lexer.tokenize(s)))



import collections.abc
import logging
from pathlib import Path
import sys
from typing import Optional


import json_streams
import json_streams.jsonlib
from sb_json_tools import jt_val
import typer

# from tabulate import tabulate
from tqdm import tqdm

from karp.foundation.commands import CommandBus
from karp import lex
from karp.utility import json_schema

# from karp.lex.domain.errors import ResourceAlreadyPublished

from karp.cliapp.utility import cli_error_handler, cli_timer
from karp.cliapp.typer_injector import inject_from_ctx

logger = logging.getLogger(__name__)


subapp = typer.Typer()


@subapp.command("add")
@cli_error_handler
@cli_timer
def add_entries_to_resource(
    ctx: typer.Context,
    resource_id: str,
    # version: Optional[int],
    data: Path,
    chunked: bool = False,
    chunk_size: int = 1000,
    user: Optional[str] = typer.Option(None),
    message: Optional[str] = typer.Option(None),
):
    bus = inject_from_ctx(CommandBus, ctx)
    user = user or "local admin"
    message = message or "imported through cli"
    entries = tqdm(json_streams.load_from_file(data), desc="Adding", unit=" entries")
    if chunked:
        cmd = lex.AddEntriesInChunks(
            resource_id=resource_id,
            chunk_size=chunk_size,
            entries=entries,
            user=user,
            message=message,
        )
    else:
        cmd = lex.AddEntries(
            resource_id=resource_id,
            entries=entries,
            user=user,
            message=message,
        )
    bus.dispatch(cmd)
    typer.echo(f"Successfully added entries to {resource_id}")


@subapp.command("import")
@cli_error_handler
@cli_timer
def import_entries_to_resource(
    ctx: typer.Context,
    resource_id: str,
    # version: Optional[int],
    data: Path,
    chunked: bool = False,
    chunk_size: int = 1000,
    user: Optional[str] = typer.Option(None),
    message: Optional[str] = typer.Option(None),
):
    bus = inject_from_ctx(CommandBus, ctx)
    user = user or "local admin"
    message = message or "imported through cli"
    entries = tqdm(json_streams.load_from_file(data), desc="Adding", unit=" entries")
    if chunked:
        cmd = lex.ImportEntriesInChunks(
            resource_id=resource_id,
            chunk_size=chunk_size,
            entries=entries,
            user=user,
            message=message,
        )
    else:
        cmd = lex.ImportEntries(
            resource_id=resource_id,
            entries=entries,
            user=user,
            message=message,
        )
    bus.dispatch(cmd)
    typer.echo(f"Successfully imported entries to {resource_id}")


@subapp.command("update")
@cli_error_handler
@cli_timer
def update_entries(resource_id: str, data: Path):
    raise NotImplementedError("")


@subapp.command("export")
@cli_error_handler
@cli_timer
def export_entries(
    ctx: typer.Context,
    resource_id: str,
    output: typer.FileBinaryWrite = typer.Option(..., "--output", "-o"),
):
    entry_views = inject_from_ctx(lex.EntryViews, ctx=ctx)
    all_entries = entry_views.all_entries(resource_id=resource_id)
    logger.debug(
        "exporting entries",
        extra={"resource_id": resource_id, "type(all_entries)": type(all_entries)},
    )
    json_streams.dump(
        (entry.dict() for entry in all_entries),
        output,
    )


class Counter(collections.abc.Generator):
    def __init__(self, sink) -> None:
        self._counter: int = 0
        self._sink = sink

    @property
    def counter(self) -> int:
        return self._counter

    def send(self, value):
        self._counter += 1
        self._sink.send(value)

    def throw(self, typ=None, val=None, tb=None):
        raise StopIteration


@subapp.command("validate")
@cli_error_handler
@cli_timer
def validate_entries(
    ctx: typer.Context,
    path: Optional[Path] = typer.Argument(None),
    config_path: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        help="resource config",
    ),
    resource_id_raw: Optional[str] = typer.Option(None, "--resource_id"),
    output: Optional[Path] = typer.Option(
        None, "--output", "-o", help="file to write to"
    ),
):
    typer.echo(f"reading from {path or 'stdin'} ...", err=True)

    if not output and path:
        output = Path(f"{path}.v6.jsonl")

    err_output = Path(f"{output}.errors.jsonl")

    if config_path and resource_id_raw:
        typer.echo("You can't provide both '--resource_id' and '--config/-c'", err=True)
        raise typer.Exit(301)

    if config_path:
        config = json_streams.jsonlib.load_from_file(config_path)
    elif resource_id_raw:
        repo = inject_from_ctx(lex.ReadOnlyResourceRepository, ctx=ctx)
        if resource := repo.get_by_resource_id(resource_id_raw):
            config = resource.config
        else:
            typer.echo(f"Can't find resource '{resource_id_raw}'", err=True)
            raise typer.Exit(302)
    else:
        typer.echo("You must provide either '--resource_id' or '--config/-c'", err=True)
        raise typer.Exit(code=300)

    schema = json_schema.create_entry_json_schema(config["fields"])

    error_code = 0

    with json_streams.sink_from_file(
        err_output, use_stderr_as_default=True
    ) as error_sink, json_streams.sink_from_file(
        output, use_stdout_as_default=True
    ) as correct_sink:
        error_counter = Counter(error_sink)
        # error_counter.send(None)
        jt_val.processing_validate(
            schema,
            tqdm(
                json_streams.load_from_file(path, use_stdin_as_default=True),
                desc="Validating",
                unit=" entries",
            ),
            on_ok=correct_sink,
            on_error=error_counter,
        )
        if error_counter.counter > 0:
            error_code = 130
            typer.echo(
                f'{error_counter.counter} entries failed validation, see "{err_output}"',
                err=True,
            )

    if error_code:
        raise typer.Exit(error_code)


def init_app(app):
    app.add_typer(subapp, name="entries")
